"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

attribute_decl = {
    'local':
"""
    // Local %(attr_type)s %(name)s
    std::vector< std::vector< %(type)s > > %(name)s;
""",
    'semiglobal':
"""
    // Semiglobal %(attr_type)s %(name)s
    std::vector< %(type)s > %(name)s ;
""",
    'global':
"""
    // Global %(attr_type)s %(name)s
    %(type)s  %(name)s ;
"""
}

attribute_cpp_init = {
    'local':
"""
        // Local %(attr_type)s %(name)s
        %(name)s = init_matrix_variable<%(type)s>(%(init)s);
""",
    'semiglobal':
"""
        // Semiglobal %(attr_type)s %(name)s
        %(name)s = init_vector_variable<%(type)s>(%(init)s);
""",
    'global':
"""
        // Global %(attr_type)s %(name)s
        %(name)s = %(init)s;
"""
}

attribute_cpp_size = {
    'local': """
        // Local %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<std::vector<%(ctype)s>>); 
        size_in_bytes += sizeof(std::vector<%(ctype)s>) * %(name)s.capacity();
        for(auto it = %(name)s.cbegin(); it != %(name)s.cend(); it++)
            size_in_bytes += (it->capacity()) * sizeof(%(ctype)s);        
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'global': """
        // Global %(attr_type)s %(name)s
        size_in_bytes += sizeof(%(ctype)s);
"""
}

attribute_cpp_delete = {
    'local': """
        // %(name)s
        for (auto it = %(name)s.begin(); it != %(name)s.end(); it++) {
            it->clear();
            it->shrink_to_fit();
        };
        %(name)s.clear();
        %(name)s.shrink_to_fit();
""",
    'semiglobal': """
        // %(name)s
        %(name)s.clear();
        %(name)s.shrink_to_fit();
""",
    'global': ""
}

cpp_11_rng = {
    'template': """#pragma omp single
{
    %(global_rng)s
    for (%(idx_type)s i = 0; i < post_rank.size(); i++) {
    %(semiglobal_rng)s
        for (%(idx_type)s j = 0; j < pre_rank[i].size(); j++) {
    %(local_rng)s
        }
    }
}
""",
    'global': """
%(rd_name)s = dist_%(rd_name)s(rng[0]);
""",
    'semiglobal': """
    %(rd_name)s[i] = dist_%(rd_name)s(rng[0]);
""",
    'local': """
        %(rd_name)s[i][j] = dist_%(rd_name)s(rng[0]);
"""
}

delay = {
    # A single value for all synapses    
    'uniform': {
        'declare': """
    // Uniform delay
    int delay ;

    int get_delay() { return delay; }
    int get_dendrite_delay(int idx) { return delay; }
    void set_delay(int delay) { this->delay = delay; }
""",
        'init': """
        delay = delays[0][0];
""",
        'reset': ""
    },
    # An individual value for each synapse
    'nonuniform_rate_coded': {
        'declare': """
    std::vector<std::vector<int>> delay;
    int max_delay;

    std::vector<std::vector<int>> get_delay() { return delay; }
    std::vector<int> get_dendrite_delay(int idx) {
        if (idx < delay.size()) {
            return delay[idx];
        } else {
            std::cerr << "ProjStruct%(id)s::get_dendrite_delay(): invalid idx used." << std::endl;
            return std::vector<int>();
        }
    }
    void set_delay(std::vector<std::vector<int>> delay) {
        this->delay = delay;
    }
    int get_max_delay() { return max_delay; }
    void set_max_delay(int max_delay) { this->max_delay = max_delay; }
""",
        'init': """
    delay = init_matrix_variable<int>(1);
    update_matrix_variable_all<int>(delay, delays);

    max_delay = %(pre_prefix)smax_delay;
""",
        'reset': ""
    },
    # An individual value for each synapse and a
    # buffer for spike events
    'nonuniform_spiking': {
        'declare': """
    std::vector<std::vector<int>> delay;
    int max_delay;
    int idx_delay;
    std::vector< std::vector< std::vector< int > > > _delayed_spikes;

    std::vector<std::vector<int>> get_delay() { return get_matrix_variable_all<int>(delay); }
    void set_delay(std::vector<std::vector<int>> value) { update_matrix_variable_all<int>(delay, value); }
    std::vector<int> get_dendrite_delay(int lil_idx) { return get_matrix_variable_row<int>(delay, lil_idx); }
    int get_max_delay() { return max_delay; }
    void set_max_delay(int max_delay) { this->max_delay = max_delay; }
""",
        'init': """
    delay = init_matrix_variable<int>(1);
    update_matrix_variable_all<int>(delay, delays);

    idx_delay = 0;
    max_delay = %(pre_prefix)smax_delay ;
    _delayed_spikes = std::vector< std::vector< std::vector< int > > >(max_delay, std::vector< std::vector< int > >(post_rank.size(), std::vector< int >()) );
""",
        'reset': """
        while(!_delayed_spikes.empty()) {
            auto elem = _delayed_spikes.back();
            elem.clear();
            _delayed_spikes.pop_back();
        }

        idx_delay = 0;
        max_delay = %(pre_prefix)smax_delay ;
        _delayed_spikes = std::vector< std::vector< std::vector< int > > >(max_delay, std::vector< std::vector< int > >(post_rank.size(), std::vector< int >()) );
"""
    }
}

event_driven = {
    'declare': """
    std::vector< std::vector<long> > _last_event;
""",
    'cpp_init': """
        _last_event = init_matrix_variable<long>(-10000);
""",
    'pyx_struct': """
        vector[vector[long]] _last_event
""",
}

###############################################################
# Rate-coded continuous transmission (general case)
###############################################################
lil_summation_operation = {
    'sum' : """
%(pre_copy)s
nb_post = static_cast<%(idx_type)s>(post_rank.size());

%(omp_code)s %(omp_clause)s %(omp_schedule)s
for (%(idx_type)s i = 0; i < nb_post; i++) {
    sum = 0.0;
    nb_pre = static_cast<%(idx_type)s>(pre_rank[i].size());
    for (%(idx_type)s j = 0; j < nb_pre; j++) {
        sum += %(psp)s ;
    }
    %(post_prefix)s_sum_%(target)s%(post_index)s += sum;
}
""",
    'max': """
%(pre_copy)s
nb_post = static_cast<%(idx_type)s>(post_rank.size());

%(omp_code)s %(omp_clause)s %(omp_schedule)s
for (%(idx_type)s i = 0; i < nb_post; i++) {
    %(idx_type)s j = 0;
    sum = %(psp)s ;
    for (j = 1; j < pre_rank[i].size(); j++) {
        if(%(psp)s > sum){
            sum = %(psp)s ;
        }
    }
    %(post_prefix)s_sum_%(target)s%(post_index)s += sum;
}
""",
    'min': """
%(pre_copy)s
nb_post = static_cast<%(idx_type)s>(post_rank.size());

%(omp_code)s %(omp_clause)s %(omp_schedule)s
for (%(idx_type)s i = 0; i < nb_post; i++) {
    %(idx_type)s j = 0;
    sum = %(psp)s ;
    for (j = 1; j < pre_rank[i].size(); j++) {
        if(%(psp)s < sum){
            sum = %(psp)s ;
        }
    }
    %(post_prefix)s_sum_%(target)s%(post_index)s += sum;
}
""",
    'mean': """
%(pre_copy)s
nb_post = static_cast<%(idx_type)s>(post_rank.size());

%(omp_code)s %(omp_clause)s %(omp_schedule)s
for (%(idx_type)s i = 0; i < nb_post; i++) {
    sum = 0.0 ;
    for(%(idx_type)s j = 0; j < pre_rank[i].size(); j++) {
        sum += %(psp)s ;
    }
    %(post_prefix)s_sum_%(target)s%(post_index)s += sum / static_cast<%(float_prec)s>(pre_rank[i].size());
}
"""
}

###############################################################
# Rate-coded continuous transmission using SIMD instructions
# and a single weight
###############################################################
lil_summation_operation_sse_single_weight = {
    'sum' : {
        'double': """
    #ifdef __SSE4_1__
        if (_transmission && %(post_prefix)s_active) {
            %(idx_type)s _s, _stop;
            double _tmp_sum[2];
            double* __restrict__ _pre_r = %(get_r)s;
            %(idx_type)s nb_post = static_cast<%(idx_type)s>(post_rank.size());

            #pragma omp for firstprivate(w)
            for (%(idx_type)s i = 0; i < nb_post; i++) {
                %(idx_type)s* __restrict__ _idx = pre_rank[i].data();
                _stop = pre_rank[i].size();

                __m128d _tmp_reg_sum = _mm_set1_pd(0.0);
                _s = 0;
                for (; _s+8 < _stop; _s+=8) {
                    __m128d _tmp_r = _mm_set_pd(_pre_r[_idx[_s+1]], _pre_r[_idx[_s]]);
                    __m128d _tmp_r2 = _mm_set_pd(_pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]]);
                    __m128d _tmp_r3 = _mm_set_pd(_pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]]);
                    __m128d _tmp_r4 = _mm_set_pd(_pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]]);

                    _tmp_reg_sum = _mm_add_pd(_tmp_reg_sum, _tmp_r);
                    _tmp_reg_sum = _mm_add_pd(_tmp_reg_sum, _tmp_r2);
                    _tmp_reg_sum = _mm_add_pd(_tmp_reg_sum, _tmp_r3);
                    _tmp_reg_sum = _mm_add_pd(_tmp_reg_sum, _tmp_r4);
                }
                _mm_storeu_pd(_tmp_sum, _tmp_reg_sum);

                // partial sums
                double lsum = _tmp_sum[0] + _tmp_sum[1];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]];

                %(post_prefix)s_sum_%(target)s%(post_index)s += w * lsum;
            }
        } // active
    #endif
""",
        'float': """
    #ifdef __SSE4_1__
        if (_transmission && %(post_prefix)s_active) {
            %(idx_type)s _s, _stop;
            float _tmp_sum[4];
            float* __restrict__ _pre_r = %(get_r)s;
            %(idx_type)s nb_post = static_cast<%(idx_type)s>(post_rank.size());

            #pragma omp for firstprivate(w)
            for (%(idx_type)s i = 0; i < nb_post; i++) {
                %(idx_type)s* __restrict__ _idx = pre_rank[i].data();
                _stop = pre_rank[i].size();

                __m128 _tmp_reg_sum = _mm_set1_ps(0.0);
                _s = 0;
                for (; _s+16 < _stop; _s+=16) {
                    __m128 _tmp_r = _mm_set_ps(_pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]], _pre_r[_idx[_s+1]], _pre_r[_idx[_s]]);
                    __m128 _tmp_r2 = _mm_set_ps(_pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]], _pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]]);
                    __m128 _tmp_r3 = _mm_set_ps(_pre_r[_idx[_s+11]], _pre_r[_idx[_s+10]], _pre_r[_idx[_s+9]], _pre_r[_idx[_s+8]]);
                    __m128 _tmp_r4 = _mm_set_ps(_pre_r[_idx[_s+15]], _pre_r[_idx[_s+14]], _pre_r[_idx[_s+13]], _pre_r[_idx[_s+12]]);

                    _tmp_reg_sum = _mm_add_ps(_tmp_reg_sum, _tmp_r);
                    _tmp_reg_sum = _mm_add_ps(_tmp_reg_sum, _tmp_r2);
                    _tmp_reg_sum = _mm_add_ps(_tmp_reg_sum, _tmp_r3);
                    _tmp_reg_sum = _mm_add_ps(_tmp_reg_sum, _tmp_r4);
                }
                _mm_storeu_ps(_tmp_sum, _tmp_reg_sum);

                // partial sums
                float lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]];

                %(post_prefix)s_sum_%(target)s%(post_index)s += w * lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with SSE4-1 support. Please check your compiler flags ..." << std::endl;
    #endif
"""
    }
}

lil_summation_operation_avx_single_weight = {
    'sum' : {
        'double': """
    #ifdef __AVX__
        if (_transmission && pop%(id_post)s->_active) {
            double* __restrict__ _pre_r = %(get_r)s;
            %(idx_type)s nb_post = static_cast<%(idx_type)s>(post_rank.size());

            #pragma omp for firstprivate(w)
            for (%(idx_type)s i = 0; i < nb_post; i++) {
                %(idx_type)s* __restrict__ _idx = pre_rank[i].data();
                %(idx_type)s _s = 0;
                %(idx_type)s _stop = static_cast<%(idx_type)s>(pre_rank[i].size());
                double _tmp_sum[4];
                __m256d _tmp_reg_sum = _mm256_setzero_pd();

                for (; (_s+8) < _stop; _s+=8) {
                    __m256d _tmp_r = _mm256_set_pd(
                        _pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]], _pre_r[_idx[_s+1]], _pre_r[_idx[_s]]
                    );
                    __m256d _tmp_r2 = _mm256_set_pd(
                        _pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]], _pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]]
                    );

                    _tmp_reg_sum = _mm256_add_pd(_tmp_reg_sum, _tmp_r);
                    _tmp_reg_sum = _mm256_add_pd(_tmp_reg_sum, _tmp_r2);
                }
                _mm256_storeu_pd(_tmp_sum, _tmp_reg_sum);

                // partial sums
                double lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]];

                %(post_prefix)s_sum_%(target)s%(post_index)s += w * lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with AVX support. Please check your compiler flags ..." << std::endl;
    #endif
    """,
        'float': """
    #ifdef __AVX__
        if (_transmission && %(post_prefix)s_active) {
            %(idx_type)s _s, _stop;
            float _tmp_sum[8];
            float* __restrict__ _pre_r = %(get_r)s;
            %(idx_type)s nb_post = static_cast<%(idx_type)s>(post_rank.size());

            #pragma omp for firstprivate(w)
            for (%(idx_type)s i = 0; i < nb_post; i++) {
                %(idx_type)s* __restrict__ _idx = pre_rank[i].data();
                _stop = pre_rank[i].size();

                __m256 _tmp_reg_sum = _mm256_set1_ps(0.0);
                __m256 _tmp_w = _mm256_set1_ps(w);
                __m256 _tmp_w2 = _mm256_set1_ps(w);

                _s = 0;
                for (; _s+16 < _stop; _s+=16) {
                    __m256 _tmp_r = _mm256_set_ps(
                        _pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]], _pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]],
                        _pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]], _pre_r[_idx[_s+1]], _pre_r[_idx[_s]]
                    );
                    __m256 _tmp_r2 = _mm256_set_ps(
                        _pre_r[_idx[_s+15]], _pre_r[_idx[_s+14]], _pre_r[_idx[_s+13]], _pre_r[_idx[_s+12]],
                        _pre_r[_idx[_s+11]], _pre_r[_idx[_s+10]], _pre_r[_idx[_s+9]], _pre_r[_idx[_s+8]]
                    );

                    _tmp_reg_sum = _mm256_add_ps(_tmp_reg_sum, _tmp_r);
                    _tmp_reg_sum = _mm256_add_ps(_tmp_reg_sum, _tmp_r2);
                }
                _mm256_storeu_ps(_tmp_sum, _tmp_reg_sum);

                // partial sums
                float lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3] + _tmp_sum[4] + _tmp_sum[5] + _tmp_sum[6] + _tmp_sum[7];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]];

                %(post_prefix)s_sum_%(target)s%(post_index)s += w * lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with AVX support. Please check your compiler flags ..." << std::endl;
    #endif
    """
    }
}

continuous_transmission_avx512_single_weight = {
    'sum' : {
        'double': """
    #ifdef __AVX512F__
        if (_transmission && pop%(id_post)s->_active) {
            double _tmp_sum[8];
            double* __restrict__ _pre_r = %(get_r)s;

            #pragma omp for firstprivate(w)
            for (%(idx_type)s i = 0; i < post_rank.size(); i++) {
                %(idx_type)s rk_post = post_rank[i];
                %(idx_type)s* __restrict__ _idx = pre_rank[i].data();
                %(idx_type)s _s = 0;
                %(idx_type)s _stop = static_cast<%(idx_type)s>(pre_rank[i].size());

                __m512d _tmp_reg_sum = _mm512_setzero_pd();

                for (; (_s+8) < _stop; _s+=8) {
                    __m512d _tmp_r = _mm512_set_pd(
                        _pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]], _pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]],
                        _pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]], _pre_r[_idx[_s+1]], _pre_r[_idx[_s]]
                    );

                    _tmp_reg_sum = _mm512_add_pd(_tmp_reg_sum, _tmp_r);
                }
                _mm512_storeu_pd(_tmp_sum, _tmp_reg_sum);

                // partial sums
                double lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3] + _tmp_sum[4] + _tmp_sum[5] + _tmp_sum[6] + _tmp_sum[7];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]];

                pop%(id_post)s._sum_%(target)s%(post_index)s +=  w * lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with AVX-512 support. Please check your compiler flags ..." << std::endl;
    #endif
    """,
        'float': """
    #ifdef __AVX512F__
        if (_transmission && pop%(id_post)s->_active) {
            float _tmp_sum[16];
            float* __restrict__ _pre_r = %(get_r)s;

            #pragma omp for firstprivate(w)
            for (%(idx_type)s i = 0; i < post_rank.size(); i++) {
                %(idx_type)s rk_post = post_rank[i];
                %(idx_type)s* __restrict__ _idx = pre_rank[i].data();
                %(idx_type)s _s = 0;
                %(idx_type)s _stop = static_cast<%(idx_type)s>(pre_rank[i].size());
                __m512 _tmp_reg_sum = _mm512_setzero_ps();

                for (; (_s+16) < _stop; _s+=16) {
                    __m512 _tmp_r = _mm512_set_ps(
                        _pre_r[_idx[_s+15]], _pre_r[_idx[_s+14]], _pre_r[_idx[_s+13]], _pre_r[_idx[_s+12]],
                        _pre_r[_idx[_s+11]], _pre_r[_idx[_s+10]], _pre_r[_idx[_s+9]], _pre_r[_idx[_s+8]],
                        _pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]], _pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]],
                        _pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]], _pre_r[_idx[_s+1]], _pre_r[_idx[_s]]
                    );

                    _tmp_reg_sum = _mm512_add_ps(_tmp_reg_sum, _tmp_r);
                }
                _mm512_storeu_ps(_tmp_sum, _tmp_reg_sum);

                // partial sums
                float lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3] + _tmp_sum[4] + _tmp_sum[5] + _tmp_sum[6] + _tmp_sum[7] + _tmp_sum[8] + _tmp_sum[9] + _tmp_sum[10] + _tmp_sum[11] + _tmp_sum[12] + _tmp_sum[13] + _tmp_sum[14] + _tmp_sum[15];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]];

                pop%(id_post)s._sum_%(target)s%(post_index)s +=  w * lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with AVX-512 support. Please check your compiler flags ..." << std::endl;
    #endif
    """
    }
}

###############################################################
# Rate-coded continuous transmission using SIMD instructions
###############################################################
continuous_transmission_sse = {
    'sum' : {
        'double': """
    #ifdef __SSE4_1__
        if (_transmission && pop%(id_post)s->_active) {
            double* __restrict__ _pre_r = %(get_r)s;
            %(idx_type)s nb_post = static_cast<%(idx_type)s>(post_rank.size());

            #pragma omp for
            for (%(idx_type)s i = 0; i < nb_post; i++) {
                %(idx_type)s* __restrict__ _idx = pre_rank[i].data();
                double* __restrict__ _w = w[i].data();
                %(idx_type)s _s = 0;
                %(idx_type)s _stop = static_cast<%(idx_type)s>(pre_rank[i].size());
                double _tmp_sum[2];
                __m128d _tmp_reg_sum = _mm_setzero_pd();

                for (; (_s+8) < _stop; _s+=8) {
                    __m128d _tmp_r = _mm_set_pd(_pre_r[_idx[_s+1]], _pre_r[_idx[_s]]);
                    __m128d _tmp_r2 = _mm_set_pd(_pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]]);
                    __m128d _tmp_r3 = _mm_set_pd(_pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]]);
                    __m128d _tmp_r4 = _mm_set_pd(_pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]]);

                    __m128d _tmp_w = _mm_loadu_pd(&_w[_s]);
                    __m128d _tmp_w2 = _mm_loadu_pd(&_w[_s+2]);
                    __m128d _tmp_w3 = _mm_loadu_pd(&_w[_s+4]);
                    __m128d _tmp_w4 = _mm_loadu_pd(&_w[_s+6]);

                    _tmp_reg_sum = _mm_add_pd(_tmp_reg_sum, _mm_mul_pd(_tmp_r, _tmp_w));
                    _tmp_reg_sum = _mm_add_pd(_tmp_reg_sum, _mm_mul_pd(_tmp_r2, _tmp_w2));
                    _tmp_reg_sum = _mm_add_pd(_tmp_reg_sum, _mm_mul_pd(_tmp_r3, _tmp_w3));
                    _tmp_reg_sum = _mm_add_pd(_tmp_reg_sum, _mm_mul_pd(_tmp_r4, _tmp_w4));
                }
                _mm_storeu_pd(_tmp_sum, _tmp_reg_sum);

                // partial sums
                double lsum = _tmp_sum[0] + _tmp_sum[1];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]] * _w[_s];

                pop%(id_post)s._sum_%(target)s%(post_index)s += lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with SSE4-1 support. Please check your compiler flags ..." << std::endl;
    #endif
    """,
        'float': """
    #ifdef __SSE4_1__
        if (_transmission && pop%(id_post)s->_active) {
            , _stop;
            float _tmp_sum[4];
            float* __restrict__ _pre_r = %(get_r)s;
            %(idx_type)s nb_post = static_cast<%(idx_type)s>(post_rank.size());

            #pragma omp for
            for (%(idx_type)s i = 0; i < nb_post; i++) {
                %(idx_type)s* __restrict__ _idx = pre_rank[i].data();
                float* __restrict__ _w = w[i].data();
                %(idx_type)s _s = 0;
                %(idx_type)s _stop = static_cast<%(idx_type)s>(pre_rank[i].size());
                __m128 _tmp_reg_sum = _mm_setzero_pd();

                for (; (_s+16) < _stop; _s+=16) {
                    __m128 _tmp_r = _mm_set_ps(
                        _pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]], _pre_r[_idx[_s+1]], _pre_r[_idx[_s]]
                    );
                    __m128 _tmp_r = _mm_set_ps(
                        _pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]], _pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]],
                    );
                    __m128 _tmp_r = _mm_set_ps(
                        _pre_r[_idx[_s+11]], _pre_r[_idx[_s+10]], _pre_r[_idx[_s+9]], _pre_r[_idx[_s+8]]
                    );
                    __m128 _tmp_r = _mm_set_ps(
                        _pre_r[_idx[_s+15]], _pre_r[_idx[_s+14]], _pre_r[_idx[_s+13]], _pre_r[_idx[_s+12]],
                    );

                    __m128 _tmp_w = _mm_loadu_ps(&_w[_s]);
                    __m128 _tmp_w2 = _mm_loadu_ps(&_w[_s+4]);
                    __m128 _tmp_w3 = _mm_loadu_ps(&_w[_s+8]);
                    __m128 _tmp_w3 = _mm_loadu_ps(&_w[_s+12]);

                    _tmp_reg_sum = _mm_add_ps(_tmp_reg_sum, _mm_mul_ps(_tmp_r, _tmp_w));
                    _tmp_reg_sum = _mm_add_ps(_tmp_reg_sum, _mm_mul_ps(_tmp_r2, _tmp_w2));
                    _tmp_reg_sum = _mm_add_ps(_tmp_reg_sum, _mm_mul_ps(_tmp_r3, _tmp_w3));
                    _tmp_reg_sum = _mm_add_ps(_tmp_reg_sum, _mm_mul_ps(_tmp_r4, _tmp_w4));
                }
                _mm_storeu_ps(_tmp_sum, _tmp_reg_sum);

                // partial sums
                float lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]] * _w[_s];

                pop%(id_post)s._sum_%(target)s%(post_index)s += lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with AVX support. Please check your compiler flags ..." << std::endl;
    #endif
    """
    }
}

continuous_transmission_avx = {
    'sum' : {
        'double': """
    #ifdef __AVX__
        if (_transmission && %(post_prefix)s_active) {
            %(idx_type)s _s, _stop;
            double _tmp_sum[4];
            double* __restrict__ _pre_r = %(get_r)s;
            %(idx_type)s nb_post = static_cast<%(idx_type)s>(post_rank.size());

            #pragma omp for
            for (%(idx_type)s i = 0; i < nb_post; i++) {
                %(idx_type)s* __restrict__ _idx = pre_rank[i].data();
                double* __restrict__ _w = w[i].data();

                _stop = pre_rank[i].size();

                __m256d _tmp_reg_sum = _mm256_set1_pd(0.0);

                _s = 0;
                for (; _s+8 < _stop; _s+=8) {
                    __m256d _tmp_r = _mm256_set_pd(
                        _pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]], _pre_r[_idx[_s+1]], _pre_r[_idx[_s]]
                    );
                    __m256d _tmp_r2 = _mm256_set_pd(
                        _pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]], _pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]]
                    );

                    __m256d _tmp_w = _mm256_loadu_pd(&_w[_s]);
                    __m256d _tmp_w2 = _mm256_loadu_pd(&_w[_s+4]);

                    _tmp_reg_sum = _mm256_add_pd(_tmp_reg_sum, _mm256_mul_pd(_tmp_r, _tmp_w));
                    _tmp_reg_sum = _mm256_add_pd(_tmp_reg_sum, _mm256_mul_pd(_tmp_r2, _tmp_w2));
                }
                _mm256_storeu_pd(_tmp_sum, _tmp_reg_sum);

                // partial sums
                double lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]] * _w[_s];

                %(post_prefix)s_sum_%(target)s%(post_index)s += lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with AVX support. Please check your compiler flags ..." << std::endl;
    #endif
    """,
        'float': """
    #ifdef __AVX__
        if (_transmission && %(post_prefix)s_active) {
            %(idx_type)s _s, _stop;
            float _tmp_sum[8];
            float* __restrict__ _pre_r = %(get_r)s;
            %(idx_type)s nb_post = static_cast<%(idx_type)s>(post_rank.size());

            #pragma omp for
            for (%(idx_type)s i = 0; i < nb_post; i++) {
                int* __restrict__ _idx = pre_rank[i].data();
                float* __restrict__ _w = w[i].data();

                _stop = pre_rank[i].size();
                __m256 _tmp_reg_sum = _mm256_set1_ps(0.0);

                _s = 0;
                for (; _s+16 < _stop; _s+=16) {
                    __m256 _tmp_r = _mm256_set_ps(
                        _pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]], _pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]],
                        _pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]], _pre_r[_idx[_s+1]], _pre_r[_idx[_s]]
                    );
                    __m256 _tmp_r2 = _mm256_set_ps(
                        _pre_r[_idx[_s+15]], _pre_r[_idx[_s+14]], _pre_r[_idx[_s+13]], _pre_r[_idx[_s+12]],
                        _pre_r[_idx[_s+11]], _pre_r[_idx[_s+10]], _pre_r[_idx[_s+9]], _pre_r[_idx[_s+8]]
                    );

                    __m256 _tmp_w = _mm256_loadu_ps(&_w[_s]);
                    __m256 _tmp_w2 = _mm256_loadu_ps(&_w[_s+8]);

                    _tmp_reg_sum = _mm256_add_ps(_tmp_reg_sum, _mm256_mul_ps(_tmp_r, _tmp_w));
                    _tmp_reg_sum = _mm256_add_ps(_tmp_reg_sum, _mm256_mul_ps(_tmp_r2, _tmp_w2));
                }
                _mm256_storeu_ps(_tmp_sum, _tmp_reg_sum);

                // partial sums
                float lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3] + _tmp_sum[4] + _tmp_sum[5] + _tmp_sum[6] + _tmp_sum[7];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]] * _w[_s];

                %(post_prefix)s_sum_%(target)s%(post_index)s += lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with AVX support. Please check your compiler flags ..." << std::endl;
    #endif
    """
    }
}

continuous_transmission_avx512 = {
    'sum' : {
        'double': """
    #ifdef __AVX512F__
        if (_transmission && pop%(id_post)s->_active) {
            double* __restrict__ _pre_r = %(get_r)s;
            double _tmp_sum[8];

            #pragma omp for
            for (%(idx_type)s i = 0; i < post_rank.size(); i++) {
                %(idx_type)s rk_post = post_rank[i];
                %(idx_type)s* __restrict__ _idx = pre_rank[i].data();
                %(idx_type)s _s = 0;
                %(idx_type)s _stop = pre_rank[i].size();
                double* __restrict__ _w = w[i].data();
		        __m512d _tmp_reg_sum = _mm512_setzero_pd();

                for (; (_s+8) < _stop; _s+=8) {
                    __m512d _tmp_r = _mm512_set_pd(
                        _pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]], _pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]],
                        _pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]], _pre_r[_idx[_s+1]], _pre_r[_idx[_s]]
                    );

                    __m512d _tmp_w = _mm512_loadu_pd(&_w[_s]);

                    _tmp_reg_sum = _mm512_add_pd(_tmp_reg_sum, _mm512_mul_pd(_tmp_r, _tmp_w));
                }
                _mm512_storeu_pd(_tmp_sum, _tmp_reg_sum);

                // partial sums
                double lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3] + _tmp_sum[4] + _tmp_sum[5] + _tmp_sum[6] + _tmp_sum[7];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]] * _w[_s];

                pop%(id_post)s._sum_%(target)s%(post_index)s +=  lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with AVX-512 support. Please check your compiler flags ..." << std::endl;
    #endif
    """,
        'float': """
    #ifdef __AVX512F__
        if (_transmission && pop%(id_post)s->_active) {
            float _tmp_sum[16];
            float* __restrict__ _pre_r = %(get_r)s;

            #pragma omp for
            for (%(idx_type)s i = 0; i < post_rank.size(); i++) {
                %(idx_type)s rk_post = post_rank[i];
                %(idx_type)s* __restrict__ _idx = pre_rank[i].data();
                %(idx_type)s _s = 0;
                %(idx_type)s _stop = pre_rank[i].size();
                float* __restrict__ _w = w[i].data();
		        __m512 _tmp_reg_sum = _mm512_setzero_ps();

                for (; (_s+16) < _stop; _s+=16) {
                    __m512 _tmp_r = _mm512_set_ps(
                        _pre_r[_idx[_s+15]], _pre_r[_idx[_s+14]], _pre_r[_idx[_s+13]], _pre_r[_idx[_s+12]],
                        _pre_r[_idx[_s+11]], _pre_r[_idx[_s+10]], _pre_r[_idx[_s+9]], _pre_r[_idx[_s+8]],
                        _pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]], _pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]],
                        _pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]], _pre_r[_idx[_s+1]], _pre_r[_idx[_s]]
                    );

                    __m512 _tmp_w = _mm512_loadu_ps(&_w[_s]);

                    _tmp_reg_sum = _mm512_add_ps(_tmp_reg_sum, _mm512_mul_ps(_tmp_r, _tmp_w));
                }
                _mm512_storeu_ps(_tmp_sum, _tmp_reg_sum);

                // partial sums
                float lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3] + _tmp_sum[4] + _tmp_sum[5] + _tmp_sum[6] + _tmp_sum[7] + _tmp_sum[8] + _tmp_sum[9] + _tmp_sum[10] + _tmp_sum[11] + _tmp_sum[12] + _tmp_sum[13] + _tmp_sum[14] + _tmp_sum[15];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]] * _w[_s];

                pop%(id_post)s._sum_%(target)s%(post_index)s +=  lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with AVX-512 support. Please check your compiler flags ..." << std::endl;
    #endif
    """
    }
}

###############################################################
# Rate-coded synaptic plasticity
###############################################################
update_variables = {
    'local': """
// Check periodicity
if(_transmission && _update && %(post_prefix)s_active && ( (t - _update_offset)%%_update_period == 0L) ){
    // Global variables
    %(global)s
    // Local variables

    #pragma omp for private(rk_post, rk_pre) firstprivate(dt)
    for(int i = 0; i < post_rank.size(); i++){
        rk_post = post_rank[i]; // Get postsynaptic rank
        // Semi-global variables
        %(semiglobal)s
        // Local variables
        for(int j = 0; j < pre_rank[i].size(); j++){
            rk_pre = pre_rank[i][j]; // Get presynaptic rank
    %(local)s
        }
    }
}
""",
    'global': """
// Check periodicity
if(_transmission && _update && %(post_prefix)s_active && ( (t - _update_offset)%%_update_period == 0L)){
    // Global variables
    %(global)s

    // Local variables
    #pragma omp for
    for(int i = 0; i < post_rank.size(); i++){
        rk_post = post_rank[i]; // Get postsynaptic rank
    %(semiglobal)s
    }
}
"""
}

###############################################################
# Spiking event-driven transmission
###############################################################
spiking_summation_fixed_delay_outer_loop = """
// Event-based summation
if (_transmission && %(post_prefix)s_active){

    // Iterate over all incoming spikes (possibly delayed constantly)
    for(int _idx_j = tid; _idx_j < %(pre_array)s.size(); _idx_j += nt) {
        // Rank of the presynaptic neuron
        int rk_j = %(pre_array)s[_idx_j];

        // Find the presynaptic neuron in the inverse connectivity matrix
        auto inv_post_ptr = inv_pre_rank.find(rk_j);
        if (inv_post_ptr == inv_pre_rank.end())
            continue;

        // List of postsynaptic neurons receiving spikes from that neuron
        std::vector< std::pair<int, int> >& inv_post = inv_post_ptr->second;
        // Number of post neurons
        int nb_post = inv_post.size();

        // Iterate over connected post neurons
        for(int _idx_i = 0; _idx_i < nb_post; _idx_i++){
            // Retrieve the correct indices
            int i = inv_post[_idx_i].first;
            int j = inv_post[_idx_i].second;

            // Event-driven integration
            %(event_driven)s

            // Update conductance
            #pragma omp atomic%(g_target)s

            // Synaptic plasticity: pre-events
            %(pre_event)s
        }
    }
} // active
"""

spiking_summation_fixed_delay_inner_loop = """
// Event-based summation
if (_transmission && %(post_prefix)s_active) {

    %(spiked_array_fusion)s

    // Iterate over all incoming spikes (possibly delayed constantly)
    for (int _idx_j = 0; _idx_j < %(pre_array)s.size(); _idx_j++ ) {
        // Rank of the presynaptic neuron
        int rk_j = %(pre_array)s[_idx_j];

        // Find the presynaptic neuron in the inverse connectivity matrix
        auto inv_post_ptr = inv_pre_rank.find(rk_j);
        if (inv_post_ptr == inv_pre_rank.end())
            continue;

        // List of postsynaptic neurons receiving spikes from that neuron
        std::vector< std::pair<int, int> >& inv_post = inv_post_ptr->second;

        // Number of post neurons
        int nb_post = inv_post.size();

        // Iterate over connected post neurons
        #pragma omp for
        for (int _idx_i = 0; _idx_i < nb_post; _idx_i++){
            // Retrieve the correct indices
            int i = inv_post[_idx_i].first;
            int j = inv_post[_idx_i].second;

            // Event-driven integration
            %(event_driven)s

            // Update conductance
            %(g_target)s

            // Synaptic plasticity: pre-events
            %(pre_event)s
        }
    }
} // active
"""

# Uses a ring buffer to process non-uniform delays in spiking networks
spiking_summation_variable_delay = """
// Event-based summation
if (_transmission && %(post_prefix)s_active){

    // Iterate over the spikes emitted during the last step in the pre population
    #pragma omp for
    for(int idx_spike=0; idx_spike<%(pre_prefix)sspiked.size(); idx_spike++){

        // Get the rank of the pre-synaptic neuron which spiked
        int rk_pre = %(pre_prefix)sspiked[idx_spike];

        // Find the presynaptic neuron in the inverse connectivity matrix
        auto inv_post_ptr = inv_pre_rank.find(rk_pre);
        if (inv_post_ptr == inv_pre_rank.end())
            continue;

        // List of post neurons receiving connections from rk_pre
        std::vector< std::pair<int, int> >& rks_post = inv_post_ptr->second;

        // Iterate over the post neurons
        for(int x=0; x<rks_post.size(); x++){
            // Index of the post neuron in the connectivity matrix
            int i = rks_post[x].first ;
            // Index of the pre neuron in the connecivity matrix
            int j = rks_post[x].second ;
            // Delay of that connection
            int d = delay[i][j]-1;
            // Index in the ring buffer
            int modulo_delay = (idx_delay + d) %% max_delay;
            // Add the spike in the ring buffer
            #pragma omp critical
            {
                _delayed_spikes[modulo_delay][i].push_back(j);
            }
        }
    }

    #pragma omp barrier

    // Iterate over all post neurons having received spikes in the previous steps
    for (int i=0; i<_delayed_spikes[idx_delay].size(); i++){
        #pragma omp for
        for (int _idx_j=0; _idx_j<_delayed_spikes[idx_delay][i].size(); _idx_j++){
            // Pre-synaptic index in the connectivity matrix
            int j = _delayed_spikes[idx_delay][i][_idx_j];

            // Event-driven integration
            %(event_driven)s
            // Update conductance
            #pragma omp critical
            {
            %(g_target)s
            }
            // Synaptic plasticity: pre-events
            %(pre_event)s
        }
        // Empty the current list of the ring buffer
        #pragma omp single
        {
            _delayed_spikes[idx_delay][i].clear();
        }
    }

    #pragma omp barrier

    #pragma omp single
    {
        // Increment the index of the ring buffer
        idx_delay = (idx_delay + 1) %% max_delay;
    }
} // active
"""

spiking_post_event = """
if(_transmission && %(post_prefix)s_active){

    #pragma omp for
    for(int _idx_i = 0; _idx_i < %(post_prefix)sspiked.size(); _idx_i++){
        // In which sub matrix the neuron take place
        int rk_post = %(post_prefix)sspiked[_idx_i];

        // Find its index in the projection
        auto it = find(post_rank.begin(), post_rank.end(), rk_post);

        // Leave if the neuron is not part of the projection
        if (it==post_rank.end()) continue;

        // which position
        int i = std::distance(post_rank.begin(), it);

        // Iterate over all synapse to this neuron
        int nb_pre = pre_rank[i].size();
        for(int j = 0; j < nb_pre; j++){
%(event_driven)s
%(post_event)s
        }
    }
}
"""

######################################
### Structural plasticity
######################################
# All code templates needed for structural plasticity.
structural_plasticity = {
    'header_struct': {
        'header': """
    /*
     * Structural plasticity
     */
    bool synapse_exists(int post_idx, int pre_rk) {
        // Check if a synapse to neuron *pre_rk* exists
        return std::binary_search(pre_rank[post_idx].begin(), pre_rank[post_idx].end(), pre_rk);
    }
    void _add_at_position(int post_idx, int pos, int pre_rk, %(float_prec)s _weight, int _delay=0%(extra_args)s) {
        pre_rank[post_idx].insert(pre_rank[post_idx].begin() + pos, pre_rk);
        w[post_idx].insert(w[post_idx].begin() + pos, _weight);

        // Update additional fields
%(delay_code)s
%(add_code)s
%(spike_add)s
%(rd_add)s
    }
    void _erase_at_position(int post_idx, int pos) {
        // Update connectivty
        pre_rank[post_idx].erase(pre_rank[post_idx].begin() + pos);
        w[post_idx].erase(w[post_idx].begin() + pos);

        // Update additional fields
%(delay_remove)s
%(add_remove)s
%(spike_remove)s
%(rd_remove)s
    }
    void add_single_synapse(int post_idx, int pre, %(float_prec)s _weight, int _delay=0%(extra_args)s) {
        // Find where to put the synapse
        int idx = pre_rank[post_idx].size();
        for(int i=0; i<pre_rank[post_idx].size(); i++){
            if(pre_rank[post_idx][i] > pre){
                idx = i;
                break;
            }
        }

        // Update connectivty
        _add_at_position(post_idx, idx, pre, _weight, _delay%(extra_args_acc)s);
    }
    void add_multiple_synapses(int post_idx, std::vector<int> pre, std::vector<%(float_prec)s> _weight, std::vector<int> _delay%(extra_args_vec_decl)s) {
        int i=0;

        for (int new_item = 0; new_item < pre.size(); new_item++) {
            int idx = pre_rank[post_idx].size();

            // Find where to put the synapse. As the ranks are ordered, we can
            // scan the array without begining from the start every time
            for (; i<pre_rank[post_idx].size(); i++) {
                if(pre_rank[post_idx][i] > pre[new_item]){
                    idx = i;
                    break;
                }
            }

            _add_at_position(post_idx, idx, pre[new_item], _weight[new_item], _delay[new_item]%(extra_args_vec_acc)s);
        }
    }
    void remove_single_synapse(int post_idx, int pre) {
        // Find synapse to be removed
        int pre_idx = -1;
        for(int i=0; i<pre_rank[post_idx].size(); i++){
            if(pre_rank[post_idx][i] == pre){
                pre_idx = i;
                break;
            }
        }

        // Sanity
        if (pre_idx == -1) 
            return;

        // Update connectivity
        _erase_at_position(post_idx, pre_idx);
    };

%(inverse_connectivity_rebuild)s
""",
        'pruning': """
    // Pruning
    bool _pruning;
    int _pruning_period;
    long int _pruning_offset;
""",
        'creating': """
    // Creating
    bool _creating;
    int _creating_period;
    long int _creating_offset;
""",
        'spiking_addcode': """
        // Add the corresponding pair in inv_pre_rank
        int idx_post = 0;
        for(int i=0; i<post_rank.size(); i++){
            if(post_rank[i] == post){
                idx_post = i;
                break;
            }
        }

        // we parallelize over post, therefore it could occur that two threads access the same pre-neuron
        #pragma omp critical
        {
            inv_pre_rank[pre].push_back(std::pair<int, int>(idx_post, idx));
        }
""",
        'spiking_removecode': """
        // Remove the corresponding pair in inv_pre_rank
        int pre = pre_rank[post][idx];
        for(int i=0; i<inv_pre_rank[pre].size(); i++){
            if(inv_pre_rank[pre][i].second == idx){
                // we parallelize over post, therefore it could occur that two threads access the same pre-neuron
                #pragma omp critical
                {
                    inv_pre_rank[pre].erase(inv_pre_rank[pre].begin() + i);
                    _dirty_connectivity = true;
                }
                break;
            }
        }
""",
        'spiking_rebuild_backwardview': """    bool _dirty_connectivity = false;
    void check_and_rebuild_inverse_connectivity() {
    #ifdef _TRACE_SIMULATION_STEPS
        std::cout << "ProjStruct::check_and_rebuild_inverse_connectivity() called at time t = " << t / dt << " ms." << std::endl;
    #endif
        if (this->_dirty_connectivity) {
            inverse_connectivity_matrix();
            _dirty_connectivity = false;
        }
    }
"""
    },

    # Structural plasticity during the simulate() call
    'create': """
    // proj%(id_proj)s creating: %(eq)s
    void creating() {
        if((_creating)&&((t - _creating_offset) %% _creating_period == 0)){
        #ifdef _TRACE_SIMULATION_STEPS
            #pragma omp master
            {
            std::cout << "ProjStruct%(id_proj)s::creating() executed at time t = " << t / dt << " ms." << std::endl;
            }
        #endif
            %(proba_init)s

            #pragma omp for
            for(int i = 0; i < post_rank.size(); i++){
                int rk_post = post_rank[i];
                for(int rk_pre = 0; rk_pre < %(pre_prefix)ssize; rk_pre++){
                    if(%(condition)s){
                        // Check if the synapse exists
                        bool _exists = false;
                        for(int k=0; k<pre_rank[i].size(); k++){
                            if(pre_rank[i][k] == rk_pre){
                                _exists = true;
                                break;
                            }
                        }

                        if((!_exists)%(proba)s){
                            //std::cout << "Creating synapse between " << rk_pre << " and " << rk_post << std::endl;
                            add_single_synapse(i, rk_pre, %(weights)s%(delay)s);
                        }
                    }
                }
            }
        }
    }
""",
    'prune': """
    // proj%(id_proj)s pruning: %(eq)s
    void pruning() {
        if((_pruning)&&((t - _pruning_offset) %% _pruning_period == 0)){
        #ifdef _TRACE_SIMULATION_STEPS
            std::cout << "ProjStruct%(id_proj)s::pruning() executed at time t = " << t / dt << " ms." << std::endl;
        #endif
            %(proba_init)s

            #pragma omp for
            for(int i = 0; i < post_rank.size(); i++){
                int rk_post = post_rank[i];
                for(int j = 0; j < pre_rank[i].size();){
                    int rk_pre = pre_rank[i][j];
                    if ((%(condition)s)%(proba)s) {
                        _erase_at_position(i, j);
                    } else {
                        ++j;
                    }
                }
            }
        }
    }
"""
}

conn_templates = {
    # accessors
    'attribute_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_cpp_size': attribute_cpp_size,
    'attribute_cpp_delete': attribute_cpp_delete,
    'delay': delay,
    'event_driven': event_driven,
    'rng_update': cpp_11_rng,

    # operations
    'rate_coded_sum': lil_summation_operation,
    'vectorized_default_psp': {
        'sse': {
            'single_w': lil_summation_operation_sse_single_weight,
            'multi_w': continuous_transmission_sse
        },
        'avx': {
            'single_w': lil_summation_operation_avx_single_weight,
            'multi_w': continuous_transmission_avx
        },
        'avx512': {
            'single_w': continuous_transmission_avx512_single_weight,
            'multi_w': continuous_transmission_avx512
        }
    },
    'update_variables': update_variables,
    'spiking_sum_fixed_delay': {
        'inner_loop': spiking_summation_fixed_delay_inner_loop,
        'outer_loop': spiking_summation_fixed_delay_outer_loop
    },
    'spiking_sum_variable_delay': spiking_summation_variable_delay,
    'post_event': spiking_post_event,
    'structural_plasticity': structural_plasticity
}

conn_ids = {
    'local_index': "[i][j]",
    'semiglobal_index': '[i]',
    'global_index': '',
    'pre_index': '[pre_rank[i][j]]',
    'post_index': '[post_rank[i]]',
    'delay_nu' : '[delay[i][j]-1]', # non-uniform delay
    'delay_u' : '[delay-1]' # uniform delay
}



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.core.Neuron import Neuron

def list_standard_neurons():
    "Returns a list of standard neuron models available."
    return [
        LeakyIntegrator, 
        Izhikevich, 
        IF_curr_exp, 
        IF_cond_exp, 
        IF_curr_alpha, 
        IF_cond_alpha, 
        HH_cond_exp, 
        EIF_cond_alpha_isfa_ista, 
        EIF_cond_exp_isfa_ista,
    ]


##################
### Leaky Integrator
##################

class LeakyIntegrator(Neuron):
    r"""
    Leaky-integrator rate-coded neuron, optionally noisy.

    This simple rate-coded neuron defines an internal variable $v(t)$ which integrates the inputs $I(t)$ with a global time constant $\tau$ and a local baseline $B$. An additive noise $N(t)$ can be optionally defined: 

    $$\tau \cdot \frac{dv(t)}{dt} + v(t) = I(t) + B + N(t)$$

    The transfer function is the positive (or rectified linear ReLU) function with a threshold $T$:

    $$r(t) = (v(t) - T)^+$$

    By default, the input $I(t)$ to this neuron is `"sum(exc) - sum(inh)"`, but this can be changed by  setting the ``sum`` argument:

    ```python
    neuron = ann.LeakyIntegrator(sum="sum(ampa)")
    ```

    By default, there is no additive noise, but the ``noise`` argument can be passed with a specific distribution:

    ```python
    neuron = ann.LeakyIntegrator(noise="Normal(0.0, 1.0)")
    ```

    Equivalent code:

    ```python
    LeakyIntegrator = Neuron(
        parameters=dict(
            tau = 10.0,
            B = ann.Parameter(0.0),
            T = 0.0,
        ), 
        equations=[
            ann.Variable()'tau * dv/dt + v = sum(exc) - sum(inh) + B', method=exponential),
            'r = pos(v - T)',
        ]
    )
    ```

    :param tau: Time constant (global).
    :param B: Baseline (local).
    :param T: Threshold (global).
    :param sum: Input sums.
    """

    # For reporting
    _instantiated = []

    def __init__(self, 
                 tau:float=10.0, 
                 B:float=0.0, 
                 T:float=0.0, 
                 sum:str='sum(exc) - sum(inh)', 
                 noise:str=None) -> None:
        
        # Create the arguments
        parameters = """
            tau = %(tau)s : population
            B = %(B)s
            T = %(T)s : population
        """ % {'tau': tau, 'B': B, 'T': T}

        # Equations for the variables
        if not noise:
            noise_def = ''
        else:
            noise_def = '+ ' + noise

        equations="""
            tau * dv/dt + v = %(sum)s + B %(noise)s : exponential
            r = pos(v - T)
        """ % { 'sum' : sum, 'noise': noise_def}

        Neuron.__init__(self, 
            parameters=parameters, equations=equations,
            name="Leaky-Integrator", 
            description="Leaky-Integrator with positive transfer function and additive noise.")

        # For reporting
        self._instantiated.append(True)

##################
### Izhikevich
##################
class Izhikevich(Neuron):
    '''
    Izhikevich quadratic spiking neuron.

    > Izhikevich, E.M. (2003). *Simple Model of Spiking Neurons, IEEE Transaction on Neural Networks*, 14:6. <http://dx.doi.org/10.1109/TNN.2003.820440>
    
    The neural equations are:

    $$\\frac{dv}{dt} = 0.04 * v^2 + 5.0 * v + 140.0 - u + I$$

    $$\\frac{du}{dt} = a * (b * v - u)$$

    By default, the conductance is "g_exc - g_inh", but this can be changed by setting the `conductance` argument:

    ```python
    neuron = ann.Izhikevich(conductance='g_ampa * (1 + g_nmda) - g_gaba')
    ```

    The synapses are instantaneous, i.e the corresponding conductance is increased from the synaptic efficiency w at the time step when a spike is received.

    The ODEs are solved using the explicit Euler method.

    Equivalent code:

    ```python
    Izhikevich = ann.Neuron(
        parameters = dict(
            noise = ann.Parameter(0.0),
            a = ann.Parameter(0.02),
            b = ann.Parameter(0.2),
            c = ann.Parameter(-65.0),
            d = ann.Parameter(8.0),
            v_thresh = ann.Parameter(30.0),
            i_offset = ann.Parameter(0.0),
        ), 
        equations = [
            'I = g_exc - g_inh + noise * Normal(0.0, 1.0) + i_offset',
            ann.Variable('dv/dt = 0.04 * v^2 + 5.0 * v + 140.0 - u + I', init = -65.0),
            ann.Variable('du/dt = a * (b*v - u)', init= -13.0),
        ],
        spike = "v > v_thresh",
        reset = "v = c; u += d",
        refractory = 0.0
    )
    ```

    The default parameters are for a regular spiking (RS) neuron. They are defined as local parameters, so their value can be set at the neuron-level. Here is the neuron definition using global parameters:


    ```python
    Izhikevich = ann.Neuron(
        parameters = dict(
            noise = 0.0,
            a = 0.02,
            b = 0.2,
            c = -65.0,
            d = 8.0,
            v_thresh = 30.0,
            i_offset = 0.0,
        ), 
        equations = [
            'I = g_exc - g_inh + noise * Normal(0.0, 1.0) + i_offset',
            ann.Variable('dv/dt = 0.04 * v^2 + 5.0 * v + 140.0 - u + I', init = -65.0),
            ann.Variable('du/dt = a * (b*v - u)', init= -13.0),
        ],
        spike = "v > v_thresh",
        reset = "v = c; u += d",
        refractory = 0.0
    )
    ```

    :param a: Speed of the recovery variable
    :param b: Scaling of the recovery variable
    :param c: Reset potential.
    :param d: Increment of the recovery variable after a spike.
    :param v_thresh: Spike threshold (mV).
    :param i_offset: external current (nA).
    :param noise: Amplitude of the normal additive noise.
    :param tau_refrac: Duration of refractory period (ms).
    :param conductance: Conductances used as inputs.
    '''

    # For reporting
    _instantiated = []
    
    def __init__(self, 
        a:float=0.02, 
        b:float=0.2, 
        c:float=-65.0, 
        d:float=8.0, 
        v_thresh:float=30.0, 
        i_offset:float=0.0, 
        noise:float=0.0, 
        tau_refrac:float=0.0, 
        conductance:str="g_exc - g_inh") -> None:
        
        # Extract which targets are defined in the conductance
        #import re
        #targets = re.findall(r'g_([\w]+)', conductance)
        
        # Create the arguments
        parameters = """
            noise = %(noise)s
            a = %(a)s
            b = %(b)s
            c = %(c)s
            d = %(d)s
            v_thresh = %(v_thresh)s
            i_offset = %(i_offset)s
            tau_refrac = %(tau_refrac)s
        """ % {'a': a, 'b':b, 'c':c, 'd':d, 'v_thresh':v_thresh, 'i_offset':i_offset, 'noise':noise, 'tau_refrac':tau_refrac}
                
        # Equations for the variables
        equations="""
            I = %(conductance)s + noise * Normal(0.0, 1.0) + i_offset
            dv/dt = 0.04 * v^2 + 5.0 * v + 140.0 - u + I : init = %(c)s
            du/dt = a * (b*v - u) : init= %(u)s
        """ % { 'conductance' : conductance, 'c':c , 'u': b*c}

        spike = """
            v > v_thresh
        """
        reset = """
            v = c
            u += d
        """
        Neuron.__init__(self, 
            parameters=parameters, equations=equations, spike=spike, reset=reset, refractory='tau_refrac',
            name="Izhikevich", description="Quadratic integrate-and-fire spiking neuron with adaptation.")

        # For reporting
        self._instantiated.append(True)


##################
### IF neurons
##################
class IF_curr_exp(Neuron):
    '''
    Leaky integrate-and-fire model with fixed threshold and decaying-exponential post-synaptic current. 
    
    (Separate synaptic currents for excitatory and inhibitory synapses).

    The ODEs are solved using the exponential Euler method.
    
    Equivalent code:

    ```python
    IF_curr_exp = Neuron(
        parameters = dict(
            v_rest = ann.Parameter(-65.0),
            cm  = ann.Parameter(1.0),
            tau_m  = ann.Parameter(20.0),
            tau_syn_E = ann.Parameter(5.0),
            tau_syn_I = ann.Parameter(5.0),
            v_thresh = ann.Parameter(-50.0),
            v_reset = ann.Parameter(-65.0),
            i_offset = ann.Parameter(0.0),
        ), 
        equations = [
            ann.Variable(
                'cm * dv/dt = cm/tau_m*(v_rest -v)   + g_exc - g_inh + i_offset',
                method='exponential', init=-65.0
            ),
            ann.Variable('tau_syn_E * dg_exc/dt = - g_exc', method='exponential'),
            ann.Variable('tau_syn_I * dg_inh/dt = - g_inh', method='exponential'),
        ],
        spike = "v > v_thresh",
        reset = "v = v_reset",
        refractory = 0.0
    )
    ```

    :param v_rest:  Resting membrane potential (mV)
    :param cm: Capacity of the membrane (nF)
    :param tau_m: Membrane time constant (ms)
    :param tau_refrac: Duration of refractory period (ms)
    :param tau_syn_E: Decay time of excitatory synaptic current (ms)
    :param tau_syn_I: Decay time of inhibitory synaptic current (ms)
    :param v_thresh: Spike threshold (mV)
    :param v_reset: Reset potential after a spike (mV)
    :param i_offset: Offset current (nA)

    '''
    # For reporting
    _instantiated = []
    
    def __init__(self, 
                 v_rest:float=-65.0, cm:float=1.0, 
                 tau_m:float=20.0, tau_refrac:float=0.0, 
                 tau_syn_E:float=5.0, tau_syn_I:float=5.0, 
                 v_thresh:float=-50.0, v_reset:float=-65.0, 
                 i_offset:float=0.0):
        
        # Create the arguments
        parameters = """
            v_rest = %(v_rest)s
            cm  = %(cm)s
            tau_m  = %(tau_m)s
            tau_refrac = %(tau_refrac)s
            tau_syn_E = %(tau_syn_E)s
            tau_syn_I = %(tau_syn_I)s
            v_thresh = %(v_thresh)s
            v_reset = %(v_reset)s
            i_offset = %(i_offset)s
        """ % {'v_rest':v_rest, 'cm':cm, 'tau_m':tau_m, 'tau_refrac':tau_refrac, 
                'tau_syn_E':tau_syn_E, 'tau_syn_I':tau_syn_I, 
                'v_thresh':v_thresh, 'v_reset':v_reset, 'i_offset':i_offset}
                
        # Equations for the variables
        equations="""    
            cm * dv/dt = cm/tau_m*(v_rest -v)   + g_exc - g_inh + i_offset : exponential, init=%(v_reset)s
            tau_syn_E * dg_exc/dt = - g_exc : exponential
            tau_syn_I * dg_inh/dt = - g_inh : exponential
        """ % {'v_reset':v_reset}

        spike = """
            v > v_thresh
        """
        reset = """
            v = v_reset
        """
        Neuron.__init__(
            self, 
            parameters=parameters, 
            equations=equations, 
            spike=spike, 
            reset=reset, 
            refractory='tau_refrac',
            name="Integrate-and-Fire", 
            description="Leaky integrate-and-fire model with fixed threshold and decaying-exponential post-synaptic current.")

        # For reporting
        self._instantiated.append(True)

class IF_cond_exp(Neuron):
    '''
    Leaky integrate-and-fire model with fixed threshold and decaying-exponential post-synaptic conductance.

    The ODEs are solved using the exponential Euler method.

    Equivalent code:

    ```python
    IF_cond_exp = Neuron(
        parameters = dict(
            v_rest = ann.Parameter(-65.0),
            cm  = ann.Parameter(1.0),
            tau_m  = ann.Parameter(20.0),
            tau_syn_E = ann.Parameter(5.0),
            tau_syn_I = ann.Parameter(5.0),
            e_rev_E = ann.Parameter(0.0),
            e_rev_I = ann.Parameter(-70.0),
            v_thresh = ann.Parameter(-50.0),
            v_reset = ann.Parameter(-65.0),
            i_offset = ann.Parameter(0.0),
        ), 
        equations = [
            ann.Variable(
                'cm * dv/dt = cm/tau_m * (v_rest -v) + g_exc * (e_rev_E - v) + g_inh * (e_rev_I - v) + i_offset',
                method='exponential', init=-65.0
            ),
            ann.Variable('tau_syn_E * dg_exc/dt = - g_exc', method='exponential'),
            ann.Variable('tau_syn_I * dg_inh/dt = - g_inh', method='exponential'),
        ],
        spike = "v > v_thresh",
        reset = "v = v_reset",
        refractory = 0.0
    )
    ```


    :param v_rest:  Resting membrane potential (mV)
    :param cm: Capacity of the membrane (nF)
    :param tau_m: Membrane time constant (ms)
    :param tau_refrac: Duration of refractory period (ms)
    :param tau_syn_E: Decay time of excitatory synaptic current (ms)
    :param tau_syn_I: Decay time of inhibitory synaptic current (ms)
    :param e_rev_E: Reversal potential of excitatory conductance (mV)
    :param e_rev_I: Reversal potential of inhibitory conductance (mV)
    :param v_thresh: Spike threshold (mV)
    :param v_reset: Reset potential after a spike (mV)
    :param i_offset: Offset current (nA)
    '''
    # For reporting
    _instantiated = []
    
    def __init__(self, v_rest=-65.0, cm=1.0, tau_m=20.0, tau_refrac=0.0, 
        tau_syn_E=5.0, tau_syn_I=5.0, e_rev_E = 0.0, e_rev_I = -70.0, 
        v_thresh=-50.0, v_reset=-65.0, i_offset=0.0):
        
        # Create the arguments
        parameters = """
            v_rest = %(v_rest)s
            cm  = %(cm)s
            tau_m  = %(tau_m)s
            tau_refrac = %(tau_refrac)s
            tau_syn_E = %(tau_syn_E)s
            tau_syn_I = %(tau_syn_I)s
            v_thresh = %(v_thresh)s
            v_reset = %(v_reset)s
            i_offset = %(i_offset)s
            e_rev_E = %(e_rev_E)s 
            e_rev_I = %(e_rev_I)s
        """ % {'v_rest':v_rest, 'cm':cm, 'tau_m':tau_m, 'tau_refrac':tau_refrac, 
                'tau_syn_E':tau_syn_E, 'tau_syn_I':tau_syn_I, 'v_thresh':v_thresh, 
                'v_reset':v_reset, 'i_offset':i_offset, 'e_rev_E': e_rev_E, 'e_rev_I': e_rev_I}
                
        # Equations for the variables
        equations="""    
            cm * dv/dt = cm/tau_m*(v_rest -v)   + g_exc * (e_rev_E - v) + g_inh * (e_rev_I - v) + i_offset : exponential, init=%(v_reset)s
            tau_syn_E * dg_exc/dt = - g_exc : exponential
            tau_syn_I * dg_inh/dt = - g_inh : exponential
        """% {'v_reset':v_reset}

        spike = """
            v > v_thresh
        """
        reset = """
            v = v_reset
        """
        Neuron.__init__(self, parameters=parameters, equations=equations, 
            spike=spike, reset=reset, refractory='tau_refrac',
            name="Integrate-and-Fire", 
            description="Leaky integrate-and-fire model with fixed threshold and decaying-exponential post-synaptic conductances.")

        # For reporting
        self._instantiated.append(True)

# Alpha conductances
class IF_curr_alpha(Neuron):
    '''
    Leaky integrate-and-fire model with fixed threshold and alpha post-synaptic currents. 
    
    Separate synaptic currents for excitatory and inhibitory synapses.

    The alpha currents are calculated through a system of two linears ODEs. After a spike is received at t_spike, it peaks at t_spike + tau_syn_X, with a maximum equal to the synaptic efficiency.

    The ODEs are solved using the exponential Euler method.

    Equivalent code:

    ```python
    IF_curr_alpha = Neuron(
        parameters = dict(
            v_rest = ann.Parameter(-65.0),
            cm  = ann.Parameter(1.0),
            tau_m  = ann.Parameter(20.0),
            tau_syn_E = ann.Parameter(5.0),
            tau_syn_I = ann.Parameter(5.0),
            v_thresh = ann.Parameter(-50.0),
            v_reset = ann.Parameter(-65.0),
            i_offset = ann.Parameter(0.0),
        ), 
        equations = [
            # Scaling
            'gmax_exc = exp((tau_syn_E - dt/2.0)/tau_syn_E)',
            'gmax_inh = exp((tau_syn_I - dt/2.0)/tau_syn_I)',

            # Membrane potential
            ann.Variable(
                'cm * dv/dt = cm/tau_m*(v_rest -v)   + alpha_exc - alpha_inh + i_offset',
                method='exponential', init=-65.0),

            # Alpha-shaped conductance
            ann.Variable('tau_syn_E * dg_exc/dt = - g_exc', method='exponential'),
            ann.Variable('tau_syn_I * dg_inh/dt = - g_inh', method='exponential'),

            ann.Variable('tau_syn_E * dalpha_exc/dt = gmax_exc * g_exc - alpha_exc', method='exponential'),
            ann.Variable('tau_syn_I * dalpha_inh/dt = gmax_inh * g_inh - alpha_inh', method='exponential'),
        ],
        spike = "v > v_thresh",
        reset = "v = v_reset",
        refractory = 0.0
    )
    ```

    :param v_rest: Resting membrane potential (mV)
    :param cm: Capacity of the membrane (nF)
    :param tau_m: Membrane time constant (ms)
    :param tau_refrac: Duration of refractory period (ms)
    :param tau_syn_E: Rise time of excitatory synaptic current (ms)
    :param tau_syn_I: Rise time of inhibitory synaptic current (ms)
    :param v_thresh: Spike threshold (mV)
    :param v_reset: Reset potential after a spike (mV)
    :param i_offset: Offset current (nA)
    '''
    # For reporting
    _instantiated = []
    
    def __init__(self, v_rest=-65.0, cm=1.0, tau_m=20.0, tau_refrac=0.0, 
        tau_syn_E=5.0, tau_syn_I=5.0, v_thresh=-50.0, v_reset=-65.0, i_offset=0.0):
        
        # Create the arguments
        parameters = f"""
            v_rest = {v_rest}
            cm  = {cm}
            tau_m  = {tau_m}
            tau_refrac = {tau_refrac}
            tau_syn_E = {tau_syn_E}
            tau_syn_I = {tau_syn_I}
            v_thresh = {v_thresh}
            v_reset = {v_reset}
            i_offset = {i_offset}
        """ 
                
        # Equations for the variables
        equations="""  
            gmax_exc = exp((tau_syn_E - dt/2.0)/tau_syn_E)
            gmax_inh = exp((tau_syn_I - dt/2.0)/tau_syn_I)  
            cm * dv/dt = cm/tau_m*(v_rest -v)   + alpha_exc - alpha_inh + i_offset : exponential, init=%(v_reset)s
            tau_syn_E * dg_exc/dt = - g_exc : exponential
            tau_syn_I * dg_inh/dt = - g_inh : exponential
            tau_syn_E * dalpha_exc/dt = gmax_exc * g_exc - alpha_exc  : exponential
            tau_syn_I * dalpha_inh/dt = gmax_inh * g_inh - alpha_inh  : exponential
        """  % {'v_reset':v_reset}

        spike = """
            v > v_thresh
        """
        reset = """
            v = v_reset
        """
        Neuron.__init__(self, parameters=parameters, equations=equations, 
            spike=spike, reset=reset, refractory='tau_refrac',
            name="Integrate-and-Fire", 
            description="Leaky integrate-and-fire model with fixed threshold and alpha post-synaptic currents.")

        # For reporting
        self._instantiated.append(True)


class IF_cond_alpha(Neuron):
    '''
    Leaky integrate-and-fire model with fixed threshold and alpha post-synaptic conductance.
    
    Separate synaptic currents for excitatory and inhibitory synapses.

    The alpha currents are calculated through a system of two linears ODEs. After a spike is received at t_spike, it peaks at t_spike + tau_syn_X, with a maximum equal to the synaptic efficiency.

    The ODEs are solved using the exponential Euler method.

    Equivalent code:

    ```python
    IF_cond_alpha = Neuron(
        parameters = dict(
            v_rest = ann.Parameter(-65.0),
            cm  = ann.Parameter(1.0),
            tau_m  = ann.Parameter(20.0),
            tau_syn_E = ann.Parameter(5.0),
            tau_syn_I = ann.Parameter(5.0),
            v_thresh = ann.Parameter(-50.0),
            v_reset = ann.Parameter(-65.0),
            i_offset = ann.Parameter(0.0),
        ), 
        equations = [
            # Scaling
            'gmax_exc = exp((tau_syn_E - dt/2.0)/tau_syn_E)',
            'gmax_inh = exp((tau_syn_I - dt/2.0)/tau_syn_I)',

            # Membrane potential
            ann.Variable(
                'cm * dv/dt = cm/tau_m*(v_rest -v) + alpha_exc  * (e_rev_E - v) + alpha_inh * (e_rev_I - v) + i_offset',
                method='exponential', init=-65.0),

            # Alpha-shaped conductance
            ann.Variable('tau_syn_E * dg_exc/dt = - g_exc', method='exponential'),
            ann.Variable('tau_syn_I * dg_inh/dt = - g_inh', method='exponential'),

            ann.Variable('tau_syn_E * dalpha_exc/dt = gmax_exc * g_exc - alpha_exc', method='exponential'),
            ann.Variable('tau_syn_I * dalpha_inh/dt = gmax_inh * g_inh - alpha_inh', method='exponential'),
        ],
        spike = "v > v_thresh",
        reset = "v = v_reset",
        refractory = 0.0
    )
    ```


    :param v_rest: Resting membrane potential (mV)
    :param cm: Capacity of the membrane (nF)
    :param tau_m: Membrane time constant (ms)
    :param tau_refrac: Duration of refractory period (ms)
    :param tau_syn_E: Rise time of excitatory synaptic current (ms)
    :param tau_syn_I: Rise time of inhibitory synaptic current (ms)
    :param v_thresh: Spike threshold (mV)
    :param v_reset: Reset potential after a spike (mV)
    :param i_offset: Offset current (nA)

    '''
    # For reporting
    _instantiated = []
    
    def __init__(self, v_rest=-65.0, cm=1.0, tau_m=20.0, tau_refrac=0.0, 
        tau_syn_E=5.0, tau_syn_I=5.0, e_rev_E = 0.0, e_rev_I = -70.0, 
        v_thresh=-50.0, v_reset=-65.0, i_offset=0.0):
        
        # Create the arguments
        parameters = """
            v_rest = %(v_rest)s
            cm  = %(cm)s
            tau_m  = %(tau_m)s
            tau_refrac = %(tau_refrac)s
            tau_syn_E = %(tau_syn_E)s
            tau_syn_I = %(tau_syn_I)s
            v_thresh = %(v_thresh)s
            v_reset = %(v_reset)s
            i_offset = %(i_offset)s
            e_rev_E = %(e_rev_E)s 
            e_rev_I = %(e_rev_I)s
        """ % {'v_rest':v_rest, 'cm':cm, 'tau_m':tau_m, 'tau_refrac':tau_refrac, 
                'tau_syn_E':tau_syn_E, 'tau_syn_I':tau_syn_I, 'v_thresh':v_thresh, 
                'v_reset':v_reset, 'i_offset':i_offset, 'e_rev_E': e_rev_E, 'e_rev_I': e_rev_I}
                
        # Equations for the variables
        equations="""    
            gmax_exc = exp((tau_syn_E - dt/2.0)/tau_syn_E)
            gmax_inh = exp((tau_syn_I - dt/2.0)/tau_syn_I)
            cm * dv/dt = cm/tau_m*(v_rest -v)   + alpha_exc * (e_rev_E - v) + alpha_inh * (e_rev_I - v) + i_offset : exponential, init=%(v_reset)s
            tau_syn_E * dg_exc/dt = - g_exc : exponential
            tau_syn_I * dg_inh/dt = - g_inh : exponential
            tau_syn_E * dalpha_exc/dt = gmax_exc * g_exc - alpha_exc  : exponential
            tau_syn_I * dalpha_inh/dt = gmax_inh * g_inh - alpha_inh  : exponential
        """ % {'v_reset': v_reset}

        spike = """
            v > v_thresh
        """
        reset = """
            v = v_reset
        """
        Neuron.__init__(self, parameters=parameters, equations=equations, 
            spike=spike, reset=reset, refractory='tau_refrac',
            name="Integrate-and-Fire", 
            description="Leaky integrate-and-fire model with fixed threshold and alpha post-synaptic conductances.")

        # For reporting
        self._instantiated.append(True)

##################
### EIF neurons
##################

class EIF_cond_exp_isfa_ista(Neuron):
    '''
    Exponential integrate-and-fire neuron with spike triggered and sub-threshold adaptation currents (isfa, ista reps.), decaying-exponential post-synaptic conductances.
     
    Definition according to:

    > Brette R and Gerstner W (2005) Adaptive Exponential Integrate-and-Fire Model as an Effective Description of Neuronal Activity. J Neurophysiol 94:3637-3642

    Equivalent code:

    ```python

    EIF_cond_exp_isfa_ista = Neuron(
        parameters = dict(
            v_rest = ann.Parameter(-70.6),
            cm = ann.Parameter(0.281), 
            tau_m = ann.Parameter(9.3667), 
            tau_syn_E = ann.Parameter(5.0),
            tau_syn_I = ann.Parameter(5.0), 
            e_rev_E = ann.Parameter(0.0),
            e_rev_I = ann.Parameter(-80.0),
            tau_w = ann.Parameter(144.0), 
            a = ann.Parameter(4.0),
            b = ann.Parameter(0.0805),
            i_offset = ann.Parameter(0.0),
            delta_T = ann.Parameter(2.0),
            v_thresh = ann.Parameter(-50.4),
            v_reset = ann.Parameter(-70.6),
            v_spike = ann.Parameter(-40.0),
        ), 
        equations = [
            # Input current
            'I = g_exc * (e_rev_E - v) + g_inh * (e_rev_I - v) + i_offset',     
            # Membrane potential     
            ann.Variable('tau_m * dv/dt = (v_rest - v +  delta_T * exp((v-v_thresh)/delta_T)) + tau_m/cm*(I - w)', init=-70.6),     
            # Recovery variable 
            'tau_w * dw/dt = a * (v - v_rest) / 1000.0 - w',       
            # Conductances
            ann.Variable('tau_syn_E * dg_exc/dt = - g_exc', method='exponential'),
            ann.Variable('tau_syn_I * dg_inh/dt = - g_inh', method='exponential'),
        ],
        spike = "v > v_spike",
        reset = """
            v = v_reset
            w += b
        """,
        refractory = 0.1
    )
    ```

    
    :param v_rest:  Resting membrane potential (mV)
    :param cm: Capacity of the membrane (nF)
    :param tau_m: Membrane time constant (ms)
    :param tau_refrac: Duration of refractory period (ms)
    :param tau_syn_E: Decay time of excitatory synaptic current (ms)
    :param tau_syn_I: Decay time of inhibitory synaptic current (ms)
    :param e_rev_E: Reversal potential for excitatory input (mV)
    :param e_rev_I: Reversal potential for inhibitory input (mv)
    :param tau_w: Time constant of the adaptation variable (ms)
    :param a: Scaling of the adaptation variable
    :param b: Increment on the adaptation variable after a spike
    :param i_offset: Offset current (nA)
    :param delta_T: Speed of the exponential (mV)
    :param v_thresh: Spike threshold for the exponential (mV)
    :param v_reset: Reset potential after a spike (mV)
    :param v_spike: Spike threshold (mV)
    '''

    # For reporting
    _instantiated = []
    
    def __init__(self, v_rest = -70.6, cm = 0.281, tau_m = 9.3667, 
        tau_refrac = 0.1, tau_syn_E = 5.0, tau_syn_I = 5.0, 
        e_rev_E = 0.0, e_rev_I = -80.0, tau_w = 144.0, 
        a = 4.0, b = 0.0805, i_offset = 0.0, delta_T = 2.0, 
        v_thresh = -50.4, v_reset = -70.6, v_spike = -40.0):
        
        # Create the arguments
        parameters = """
            v_rest     = %(v_rest)s
            cm         = %(cm)s
            tau_m      = %(tau_m)s
            tau_refrac = %(tau_refrac)s
            tau_syn_E  = %(tau_syn_E)s
            tau_syn_I  = %(tau_syn_I)s
            e_rev_E    = %(e_rev_E)s
            e_rev_I    = %(e_rev_I)s
            tau_w      = %(tau_w)s
            a          = %(a)s
            b          = %(b)s
            i_offset   = %(i_offset)s
            delta_T    = %(delta_T)s
            v_thresh   = %(v_thresh)s
            v_reset    = %(v_reset)s
            v_spike    = %(v_spike)s
        """ % {
            'v_rest'     : v_rest    ,
            'cm'         : cm        ,
            'tau_m'      : tau_m     ,
            'tau_refrac' : tau_refrac,
            'tau_syn_E'  : tau_syn_E ,
            'tau_syn_I'  : tau_syn_I ,
            'e_rev_E'    : e_rev_E   ,
            'e_rev_I'    : e_rev_I   ,
            'tau_w'      : tau_w     ,
            'a'          : a         ,
            'b'          : b         ,
            'i_offset'   : i_offset  ,
            'delta_T'    : delta_T   ,
            'v_thresh'   : v_thresh  ,
            'v_reset'    : v_reset   ,
            'v_spike'    : v_spike   ,
        }
        # Equations for the variables
        equations="""    
            I = g_exc * (e_rev_E - v) + g_inh * (e_rev_I - v) + i_offset

            tau_m * dv/dt = (v_rest - v +  delta_T * exp((v-v_thresh)/delta_T)) + tau_m/cm*(I - w) : init=%(v_reset)s

            tau_w * dw/dt = a * (v - v_rest) / 1000.0 - w 

            tau_syn_E * dg_exc/dt = - g_exc : exponential
            tau_syn_I * dg_inh/dt = - g_inh : exponential
        """ % {'v_reset': v_reset}

        spike = """
            v > v_spike
        """
        reset = """
            v = v_reset
            w += b
        """
        Neuron.__init__(self, parameters=parameters, equations=equations, 
            spike=spike, reset=reset, refractory='tau_refrac',
            name="Adaptive exponential Integrate-and-Fire", 
            description="Exponential integrate-and-fire neuron with spike triggered and sub-threshold adaptation currents (isfa, ista reps.).")

        # For reporting
        self._instantiated.append(True)

class EIF_cond_alpha_isfa_ista(Neuron):
    ''' 
    Exponential integrate-and-fire neuron with spike triggered and sub-threshold adaptation conductances (isfa, ista reps.), alpha post-synaptic conductances.
     
    Definition according to:

    > Brette R and Gerstner W (2005) Adaptive Exponential Integrate-and-Fire Model as an Effective Description of Neuronal Activity. J Neurophysiol 94:3637-3642

    Equivalent code:

    ```python

    EIF_cond_alpha_isfa_ista = Neuron(
        parameters = dict(
            v_rest = ann.Parameter(-70.6),
            cm = ann.Parameter(0.281), 
            tau_m = ann.Parameter(9.3667), 
            tau_syn_E = ann.Parameter(5.0),
            tau_syn_I = ann.Parameter(5.0), 
            e_rev_E = ann.Parameter(0.0),
            e_rev_I = ann.Parameter(-80.0),
            tau_w = ann.Parameter(144.0), 
            a = ann.Parameter(4.0),
            b = ann.Parameter(0.0805),
            i_offset = ann.Parameter(0.0),
            delta_T = ann.Parameter(2.0),
            v_thresh = ann.Parameter(-50.4),
            v_reset = ann.Parameter(-70.6),
            v_spike = ann.Parameter(-40.0),
        ), 
        equations = [
            # Scaling
            'gmax_exc = exp((tau_syn_E - dt/2.0)/tau_syn_E)',
            'gmax_inh = exp((tau_syn_I - dt/2.0)/tau_syn_I)',

            # Input current
            'I = alpha_exc * (e_rev_E - v) + alpha_inh * (e_rev_I - v) + i_offset',    

            # Membrane potential     
            ann.Variable('tau_m * dv/dt = (v_rest - v +  delta_T * exp((v-v_thresh)/delta_T)) + tau_m/cm*(I - w)', init=-70.6),   

            # Recovery variable 
            'tau_w * dw/dt = a * (v - v_rest) / 1000.0 - w',    

            # Alpha-shaped conductance
            ann.Variable('tau_syn_E * dg_exc/dt = - g_exc', method='exponential'),
            ann.Variable('tau_syn_I * dg_inh/dt = - g_inh', method='exponential'),

            ann.Variable('tau_syn_E * dalpha_exc/dt = gmax_exc * g_exc - alpha_exc', method='exponential'),
            ann.Variable('tau_syn_I * dalpha_inh/dt = gmax_inh * g_inh - alpha_inh', method='exponential'),
        ],
        spike = "v > v_spike",
        reset = """
            v = v_reset
            w += b
        """,
        refractory = 0.1
    )
    ```

    :param v_rest:  Resting membrane potential (mV)
    :param cm: Capacity of the membrane (nF)
    :param tau_m: Membrane time constant (ms)
    :param tau_refrac: Duration of refractory period (ms)
    :param tau_syn_E: Decay time of excitatory synaptic current (ms)
    :param tau_syn_I: Decay time of inhibitory synaptic current (ms)
    :param e_rev_E: Reversal potential for excitatory input (mV)
    :param e_rev_I: Reversal potential for inhibitory input (mv)
    :param tau_w: Time constant of the adaptation variable (ms)
    :param a: Scaling of the adaptation variable
    :param b: Increment on the adaptation variable after a spike
    :param i_offset: Offset current (nA)
    :param delta_T: Speed of the exponential (mV)
    :param v_thresh: Spike threshold for the exponential (mV)
    :param v_reset: Reset potential after a spike (mV)
    :param v_spike: Spike threshold (mV)
    '''
    # For reporting
    _instantiated = []
    
    def __init__(self, v_rest = -70.6, cm = 0.281, tau_m = 9.3667, 
        tau_refrac = 0.1, tau_syn_E = 5.0, tau_syn_I = 5.0, 
        e_rev_E = 0.0, e_rev_I = -80.0, tau_w = 144.0, 
        a = 4.0, b = 0.0805, i_offset = 0.0, delta_T = 2.0, 
        v_thresh = -50.4, v_reset = -70.6, v_spike = -40.0):
        
        # Create the arguments
        parameters = """
            v_rest     = %(v_rest)s
            cm         = %(cm)s
            tau_m      = %(tau_m)s
            tau_refrac = %(tau_refrac)s
            tau_syn_E  = %(tau_syn_E)s
            tau_syn_I  = %(tau_syn_I)s
            e_rev_E    = %(e_rev_E)s
            e_rev_I    = %(e_rev_I)s
            tau_w      = %(tau_w)s
            a          = %(a)s
            b          = %(b)s
            i_offset   = %(i_offset)s
            delta_T    = %(delta_T)s
            v_thresh   = %(v_thresh)s
            v_reset    = %(v_reset)s
            v_spike    = %(v_spike)s
        """ % {
            'v_rest'     : v_rest    ,
            'cm'         : cm        ,
            'tau_m'      : tau_m     ,
            'tau_refrac' : tau_refrac,
            'tau_syn_E'  : tau_syn_E ,
            'tau_syn_I'  : tau_syn_I ,
            'e_rev_E'    : e_rev_E   ,
            'e_rev_I'    : e_rev_I   ,
            'tau_w'      : tau_w     ,
            'a'          : a         ,
            'b'          : b         ,
            'i_offset'   : i_offset  ,
            'delta_T'    : delta_T   ,
            'v_thresh'   : v_thresh  ,
            'v_reset'    : v_reset   ,
            'v_spike'    : v_spike   ,
        }
        # Equations for the variables
        equations="""    

            gmax_exc = exp((tau_syn_E - dt/2.0)/tau_syn_E)
            gmax_inh = exp((tau_syn_I - dt/2.0)/tau_syn_I)
            
            I = alpha_exc * (e_rev_E - v) + alpha_inh * (e_rev_I - v) + i_offset

            tau_m * dv/dt = (v_rest - v +  delta_T * exp((v-v_thresh)/delta_T)) + tau_m/cm*(I - w) : init=%(v_reset)s

            tau_w * dw/dt = a * (v - v_rest) / 1000.0 - w 

            tau_syn_E * dg_exc/dt = - g_exc : exponential
            tau_syn_I * dg_inh/dt = - g_inh : exponential
            tau_syn_E * dalpha_exc/dt = gmax_exc * g_exc - alpha_exc  : exponential
            tau_syn_I * dalpha_inh/dt = gmax_inh * g_inh - alpha_inh  : exponential
        """ % {'v_reset': v_reset}

        spike = """
            v > v_spike
        """
        reset = """
            v = v_reset
            w += b
        """
        Neuron.__init__(self, parameters=parameters, equations=equations, 
            spike=spike, reset=reset, refractory='tau_refrac',
            name="Adaptive exponential Integrate-and-Fire", 
            description="Exponential integrate-and-fire neuron with spike triggered and sub-threshold adaptation conductances (isfa, ista reps.).")

        # For reporting
        self._instantiated.append(True)

##################
### HH
##################
class HH_cond_exp(Neuron):
    '''
    Single-compartment Hodgkin-Huxley-type neuron with transient sodium and delayed-rectifier potassium currents using the ion channel models from Traub.

    The ODEs for n, m, h and v are solved using the midpoint method, while the conductances g_exc and g_inh are solved using the exponential Euler method.

    Equivalent code:

    ```python
    HH_cond_exp = Neuron(
        parameters = dict(
            gbar_Na = 20.0
            gbar_K = 6.0
            gleak = 0.01
            cm = 0.2 
            v_offset = -63.0 
            e_rev_Na = 50.0
            e_rev_K = -90.0 
            e_rev_leak = -65.0
            e_rev_E = 0.0
            e_rev_I = -80.0 
            tau_syn_E = 0.2
            tau_syn_I = 2.0
            i_offset = 0.0
            v_thresh = 0.0
        ), 
        equations = [
            # Previous membrane potential
            'prev_v = v',

            # Voltage-dependent rate constants
            'an = 0.032 * (15.0 - v + v_offset) / (exp((15.0 - v + v_offset)/5.0) - 1.0)',
            'am = 0.32  * (13.0 - v + v_offset) / (exp((13.0 - v + v_offset)/4.0) - 1.0)',
            'ah = 0.128 * exp((17.0 - v + v_offset)/18.0)',

            'bn = 0.5   * exp ((10.0 - v + v_offset)/40.0)',
            'bm = 0.28  * (v - v_offset - 40.0) / (exp((v - v_offset - 40.0)/5.0) - 1.0)',
            'bh = 4.0/(1.0 + exp (( 10.0 - v + v_offset )) )',

            # Activation variables
            ann.Variable('dn/dt = an * (1.0 - n) - bn * n', init = 0.0, method="exponential")
            ann.Variable('dm/dt = am * (1.0 - m) - bm * m', init = 0.0, method="exponential")
            ann.Variable('dh/dt = ah * (1.0 - h) - bh * h', init = 1.0, method="exponential")

            # Membrane equation
            ann.Variable('cm * dv/dt = gleak*(e_rev_leak -v) + gbar_K * n**4 * (e_rev_K - v) + gbar_Na * m**3 * h * (e_rev_Na - v)
                            + g_exc * (e_rev_E - v) + g_inh * (e_rev_I - v) + i_offset', method="exponential", init=-65.0),

            # Exponentially-decaying conductances
            ann.Variable('tau_syn_E * dg_exc/dt = - g_exc', method="exponential")
            ann.Variable('tau_syn_I * dg_inh/dt = - g_inh', method="exponential")
        ],
        spike = "(v > v_thresh) and (prev_v <= v_thresh)",
        reset = ""
    )
    ```

    :param gbar_Na: Maximal conductance of the Sodium current.
    :param gbar_K: Maximal conductance of the Potassium current. 
    :param gleak: Conductance of the leak current (nF)  
    :param cm: Capacity of the membrane (nF)
    :param v_offset:  Threshold for the rate constants (mV)  
    :param e_rev_Na: Reversal potential for the Sodium current (mV) 
    :param e_rev_K: Reversal potential for the Potassium current (mV)  
    :param e_rev_leak: Reversal potential for the leak current (mV)   
    :param e_rev_E: Reversal potential for excitatory input (mV)  
    :param e_rev_I: Reversal potential for inhibitory input (mV)  
    :param tau_syn_E: Decay time of excitatory synaptic current (ms)  
    :param tau_syn_I: Decay time of inhibitory synaptic current (ms)   
    :param i_offset: Offset current (nA)
    :param v_thresh: Threshold for spike emission

    '''
    # For reporting
    _instantiated = []
    
    def __init__(self, gbar_Na = 20.0, gbar_K = 6.0, gleak = 0.01, cm = 0.2, 
        v_offset = -63.0, e_rev_Na = 50.0, e_rev_K = -90.0, e_rev_leak = -65.0, 
        e_rev_E = 0.0, e_rev_I = -80.0, tau_syn_E = 0.2, tau_syn_I = 2.0, 
        i_offset = 0.0, v_thresh = 0.0):

        parameters = """
        gbar_Na    = %(gbar_Na)s   
        gbar_K     = %(gbar_K)s     
        gleak      = %(gleak)s      
        cm         = %(cm)s        
        v_offset   = %(v_offset)s 
        e_rev_Na   = %(e_rev_Na)s 
        e_rev_K    = %(e_rev_K)s   
        e_rev_leak = %(e_rev_leak)s   
        e_rev_E    = %(e_rev_E)s    
        e_rev_I    = %(e_rev_I)s   
        tau_syn_E  = %(tau_syn_E)s 
        tau_syn_I  = %(tau_syn_I)s 
        i_offset   = %(i_offset)s  
        v_thresh   = %(v_thresh)s  
        """ % {
        'gbar_Na'    : gbar_Na   ,
        'gbar_K'     : gbar_K     ,
        'gleak'      : gleak      ,
        'cm'         : cm        ,
        'v_offset'   : v_offset  ,
        'e_rev_Na'   : e_rev_Na  ,
        'e_rev_K'    : e_rev_K   ,
        'e_rev_leak' : e_rev_leak   ,
        'e_rev_E'    : e_rev_E    ,
        'e_rev_I'    : e_rev_I   ,
        'tau_syn_E'  : tau_syn_E ,
        'tau_syn_I'  : tau_syn_I ,
        'i_offset'   : i_offset  ,
        'v_thresh'   : v_thresh  
        }

        equations = """
            # Previous membrane potential
            prev_v = v

            # Voltage-dependent rate constants
            an = 0.032 * (15.0 - v + v_offset) / (exp((15.0 - v + v_offset)/5.0) - 1.0)
            am = 0.32  * (13.0 - v + v_offset) / (exp((13.0 - v + v_offset)/4.0) - 1.0)
            ah = 0.128 * exp((17.0 - v + v_offset)/18.0) 

            bn = 0.5   * exp ((10.0 - v + v_offset)/40.0)
            bm = 0.28  * (v - v_offset - 40.0) / (exp((v - v_offset - 40.0)/5.0) - 1.0)
            bh = 4.0/(1.0 + exp (( 10.0 - v + v_offset )) )

            # Activation variables
            dn/dt = an * (1.0 - n) - bn * n : init = 0.0, exponential
            dm/dt = am * (1.0 - m) - bm * m : init = 0.0, exponential
            dh/dt = ah * (1.0 - h) - bh * h : init = 1.0, exponential

            # Membrane equation
            cm * dv/dt = gleak*(e_rev_leak -v) + gbar_K * n**4 * (e_rev_K - v) + gbar_Na * m**3 * h * (e_rev_Na - v) 
                + g_exc * (e_rev_E - v) + g_inh * (e_rev_I - v) + i_offset: exponential, init=%(e_rev_leak)s

            # Exponentially-decaying conductances
            tau_syn_E * dg_exc/dt = - g_exc : exponential
            tau_syn_I * dg_inh/dt = - g_inh : exponential
        """ % {'e_rev_leak': e_rev_leak}

        spike = "(v > v_thresh) and (prev_v <= v_thresh)"

        reset = ""

        Neuron.__init__(self, parameters=parameters, equations=equations, 
            spike=spike, reset=reset,
            name="Hodgkin-Huxley", 
            description="Single-compartment Hodgkin-Huxley-type neuron with transient sodium and delayed-rectifier potassium currents.")

        # For reporting
        self._instantiated.append(True)


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.intern.NetworkManager import NetworkManager
from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.generator.Monitor import BaseTemplates as RecTemplate
from ANNarchy.generator.Utils import tabify
from ANNarchy.extensions.bold import BoldMonitor

from ANNarchy.generator.Monitor import OpenMPTemplates
from ANNarchy.generator.Monitor import CUDATemplates

class MonitorGenerator(object):
    """
    Creates the required codes for recording population
    and projection data
    """
    def __init__(self, annarchy_dir, net_id):
        """
        Constructor, stores all required data for later
        following code generation step

        Parameters:

            * *annarchy_dir*: unique target directory for the generated code
              files; they are stored in 'generate' sub-folder
            * *net_id*: unique id for the current network

        """
        self._annarchy_dir = annarchy_dir
        self.net_id = net_id
        self._network = NetworkManager().get_network(self.net_id)
        self._populations = self._network.get_populations()
        self._projections = self._network.get_projections()

    def generate(self):
        """
        Generate one file "Monitor.hpp" comprising of Monitor base class and inherited
        classes for each Population/Projection.

        Templates:

            record_base_class
        """
        record_class = ""
        # We generate for each of the population/projection in the network
        # a record class, containing by default all variables. The recording
        # is then enabled or disabled.
        for pop in self._populations:
            record_class += self._pop_recorder_class(pop)

        for proj in self._projections:
            record_class += self._proj_recorder_class(proj)

        code = RecTemplate.record_base_class % {'record_classes': record_class}

        # The approach for default populations/projections is not
        # feasible for specific monitors, so we handle them extra
        for mon in NetworkManager().get_network(net_id=self.net_id).get_monitors():
            if isinstance(mon, BoldMonitor):
                mon_dict = {
                    'pop_id': mon.object.id,
                    'pop_name': mon.object.name,
                    'mon_id': mon.id,
                    'float_prec': ConfigManager().get('precision', self.net_id),
                    'var_name': mon.variables[0],
                }
                code += mon._specific_template['cpp'] % mon_dict

        # Generate header code for the analysed pops and projs
        with open(self._annarchy_dir+'/generate/net'+str(self.net_id)+'/Monitor.hpp', 'w') as ofile:
            ofile.write(code)

    def _pop_recorder_class(self, pop):
        """
        Creates population recording class code.

        Returns:

            * complete code as string

        Templates:

            omp_population, cuda_population
        """
        if ConfigManager().get('paradigm', self.net_id) == "openmp":
            template = OpenMPTemplates.omp_population
        elif ConfigManager().get('paradigm', self.net_id) == "cuda":
            template = CUDATemplates.cuda_population
        else:
            raise NotImplementedError

        tpl_code = template['template']

        init_code = ""
        recording_code = ""
        recording_target_code = ""
        struct_code = ""
        size_in_bytes = ""
        clear_all_code = ""
        clear_individual_code = ""

        # The post-synaptic potential for rate-code (weighted sum) as well
        # as the conductance variables are handled seperatly.
        target_list = []
        targets = []
        for t in pop.neuron_type.description['targets']:
            if isinstance(t, list):
                for t2 in t:
                    targets.append(t2)
            else:
                targets.append(t)
        for t in pop.targets:
            if isinstance(t, list):
                for t2 in t:
                    targets.append(t2)
            else:
                targets.append(t)
        targets = sorted(list(set(targets)))

        if pop.neuron_type.type == 'rate':
            for target in targets:
                tar_dict = {'id': pop.id, 'type' : ConfigManager().get('precision', self.net_id), 'name': '_sum_'+target}
                struct_code += template['local']['struct'] % tar_dict
                init_code += template['local']['init'] % tar_dict
                recording_target_code += template['local']['recording'] % tar_dict
                clear_all_code += "\t\tthis->clear_{name}();\n".format(name='_sum_'+target)
                clear_individual_code += template['local']['clear'] % tar_dict
        else:
            for target in targets:
                tar_dict = {'id': pop.id, 'type' : ConfigManager().get('precision', self.net_id), 'name': 'g_'+target}
                struct_code += template['local']['struct'] % tar_dict
                init_code += template['local']['init'] % tar_dict
                recording_target_code += template['local']['recording'] % tar_dict
                clear_all_code += "\t\tthis->clear_{name}();\n".format(name='g_'+target)
                clear_individual_code += template['local']['clear'] % tar_dict

                # to skip this entry in the following loop
                target_list.append('g_'+target)

        # Record global and local attributes
        attributes = []
        for var in pop.neuron_type.description['variables']:
            # Skip targets
            if var['name'] in target_list:
                continue
            # Avoid doublons
            if var['name'] in attributes:
                continue
            attributes.append(var['name'])

            ids = {
                'id': pop.id,
                'name': var['name'],
                'type': var['ctype']
            }

            struct_code += template[var['locality']]['struct'] % ids
            init_code += template[var['locality']]['init'] % ids
            recording_code += template[var['locality']]['recording'] % ids
            clear_all_code += "\t\tthis->clear_{name}();\n".format(name=ids['name'])
            clear_individual_code += template[var['locality']]['clear'] % ids
            size_in_bytes += template[var['locality']]['size_in_bytes'] % ids

        # Record spike events
        if pop.neuron_type.type == 'spike':
            base_tpl = RecTemplate.recording_spike_tpl
            rec_dict = {
                'id': pop.id,
                'type' : 'long int',
                'name': 'spike',
                'rec_target': 'spiked'
            }

            struct_code += base_tpl['struct'] % rec_dict
            init_code += base_tpl['init'] % rec_dict
            recording_code += base_tpl['record'][ConfigManager().get('paradigm', self.net_id)] % rec_dict
            size_in_bytes += base_tpl['size_in_bytes'][ConfigManager().get('paradigm', self.net_id)] % rec_dict
            clear_all_code += "\t\tthis->clear_spike();\n"

            # Record axon spike events
            if pop.neuron_type.axon_spike:
                rec_dict = {
                    'id': pop.id,
                    'type' : 'long int',
                    'name': 'axon_spike',
                    'rec_target': 'axonal'
                }

                struct_code += base_tpl['struct'] % rec_dict
                init_code += base_tpl['init'] % rec_dict
                recording_code += base_tpl['record'][ConfigManager().get('paradigm', self.net_id)] % rec_dict

        ids = {
            'id': pop.id,
            'init_code': init_code,
            'struct_code': struct_code,
            'recording_code': recording_code,
            'recording_target_code': recording_target_code,
            'size_in_bytes': tabify(size_in_bytes, 2),
            'clear_all_container_code': clear_all_code,
            'clear_individual_container_code': clear_individual_code
        }
        return tpl_code % ids

    def _proj_recorder_class(self, proj):
        """
        Generate the code for the recorder object.

        Returns:

            * complete code as string

        Templates:

            record
        """
        if ConfigManager().get('paradigm', self.net_id) == "openmp":
            template = OpenMPTemplates.omp_projection
        elif ConfigManager().get('paradigm', self.net_id) == "cuda":
            template = CUDATemplates.cuda_projection
        else:
            raise NotImplementedError

        # Specific template
        if 'monitor_class' in proj._specific_template.keys():
            return proj._specific_template['monitor_class']

        init_code = ""
        recording_code = ""
        size_in_bytes_code = ""
        clear_individual_code = ""
        clear_container_code = ""
        struct_code = ""

        attributes = []
        for var in proj.synapse_type.description['variables']:
            # Avoid doublons
            if var['name'] in attributes:
                continue
            attributes.append(var['name'])

            # Get the locality
            locality = var['locality']

            # Special case for single weights
            if var['name'] == "w" and proj._has_single_weight():
                locality = 'global'

            ids = {
                'type' : var['ctype'],
                'name': var['name']
            }

            # Get the template for the structure declaration
            struct_code += template[locality]['struct'] % ids

            # Get the initialization code
            init_code += template[locality]['init'] % ids

            # Clear one specific recorded attribtute
            clear_individual_code += template[var['locality']]['clear'] % ids

            # Clear all recorded attributes
            clear_container_code += "\tthis->clear_{name}();\n".format(name=ids['name'])

            # Memory requirement
            size_in_bytes_code += tabify(template[locality]['size_in_bytes'] % ids, 2)

            # Get the recording code
            recording_code += template[locality]['recording'] % {
                'id': proj.id,
                'type' : var['ctype'],
                'name': var['name'],
                'float_prec': ConfigManager().get('precision', self.net_id)
            }

        final_dict = {
            'id': proj.id,
            'init_code': init_code,
            'recording_code': recording_code,
            'size_in_bytes_code': size_in_bytes_code,
            'clear_container_code': clear_container_code,
            'clear_individual_code': clear_individual_code,
            'struct_code': struct_code
        }

        return template['struct'] % final_dict



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

import argparse
from ANNarchy.intern import Messages
from ANNarchy.intern import ConfigManagement

class CmdLineArgParser(object):
    """
    ANNarchy scripts can be run by several command line arguments. These are
    checked with the ArgumentParser provided by Python.

    Generally, we have a group of flags which should be set before *compile()*
    and some which are important for simulation.
    """
    def __init__(self):
        # Create parser instance
        self.setup_parser()
    
    def setup_parser(self):
        """
        We setup the list of possible command line arguments. This selection is presented if *--help* has been added to the script call.
        """
        # override the error behavior of OptionParser,
        # normally an unknwon arg would raise an exception
        self.parser = argparse.ArgumentParser(description='ANNarchy: Artificial Neural Networks architect.')

        group = self.parser.add_argument_group('General')
        group.add_argument("-c", "--clean", help="Forces recompilation.", action="store_true", default=False, dest="clean")
        group.add_argument("--prec", help="Set the floating precision used.", action="store", type=str, default=None, dest="precision")
        group.add_argument("--report", help="Create a network overview using either .tex or .md.", action="store", type=str, default=None, dest="report")

        group = self.parser.add_argument_group('Debugging')
        group.add_argument("--trace", help="Trace initialization of objects (init), function calls during simulation (simulate), or all (both).", action="store", default=None, dest="trace_calls")
        group.add_argument("-d", "--debug", help="Compilation with debug symbols and additional checks.", action="store_true", default=False, dest="debug")
        group.add_argument("-v", "--verbose", help="Shows all messages.", action="store_true", default=None, dest="verbose")

        group = self.parser.add_argument_group('Performance-related')
        group.add_argument("--auto-tuning", help="Enable automatic sparse matrix format selection.", action="store_true", default=False, dest="auto_tuning")

        group = self.parser.add_argument_group('OpenMP')
        group.add_argument("-j", "--num-threads", help="Number of threads to use.", type=int, action="store", default=None, dest="num_threads")
        group.add_argument("--visible-cores", help="Cores where the threads should be placed.", type=str, action="store", default=None, dest="visible_cores")

        group = self.parser.add_argument_group('CUDA')
        group.add_argument("--gpu", help="Enables CUDA and optionally specifies the GPU id (default: 0).", type=int, action="store", nargs='?', default=-1, const=0, dest="gpu_device")

        group = self.parser.add_argument_group('Internal')
        group.add_argument("--profile", help="Enables profiling.", action="store_true", default=None, dest="profile")
        group.add_argument("--profile-out", help="Target file for profiling data.", action="store", type=str, default=None, dest="profile_out")

    def parse_arguments_for_setup(self):
        """
        We already parse some arguments which should be known before **any** other ANNarchy call has happened.
        """
        options, _ = self.parser.parse_known_args()

        # Enables time measurement within the C++ simulation core
        if options.profile:
            ConfigManagement._update_global_config('profiling', options.profile)

        # if the parameters set on command-line they overwrite Global.config
        if options.num_threads is not None:
            ConfigManagement._update_global_config('num_threads', options.num_threads)

        if options.visible_cores is not None:
            try:
                core_list = [int(x) for x in options.visible_cores.split(",")]
                ConfigManagement._update_global_config('visible_cores', core_list)
            except:
                Messages._error("As argument for 'visible_cores' a comma-seperated list of integers is expected.")

        # Get Performance-related flags
        if options.auto_tuning:
            Messages._info("Automatic format selection is an experimental feature. We greatly appreciate bug reports.")
            ConfigManagement.setup(sparse_matrix_format="auto", sparse_matrix_storage_order="auto")

        # Get CUDA configuration
        if options.gpu_device >= 0:
            ConfigManagement._update_global_config('paradigm', 'cuda')

        # Verbose
        if options.verbose is not None:
            ConfigManagement.setup(verbose = options.verbose)

        # Trace init/simulation steps
        if options.trace_calls is not None:
            if options.trace_calls in ["init", "simulate", "both"]:
                ConfigManagement._update_global_config('trace_calls', options.trace_calls)
            else:
                raise ValueError("--trace accepts only one of the following keys: init, simulate, both")

        # Precision
        if options.precision is not None:
            ConfigManagement._update_global_config('precision', options.precision)




"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

import numpy as np

from ANNarchy.intern.SpecificPopulation import SpecificPopulation
from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.intern import Messages
from ANNarchy.core.Population import Population
from ANNarchy.core.Neuron import Neuron

class TimedArray(SpecificPopulation):
    """
    Data structure holding sequential inputs for a rate-coded network.

    The input values are stored in the (recordable) attribute `r`, without any further processing.

    By default, the firing rate of this population will iterate over the different values step by step:

    ```python
    inputs = np.array(
        [
            [1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 1]
        ]
    )

    net = ann.Network()
    inp = net.create(ann.TimedArray(rates=inputs))
    pop = net.create(10, ann.LeakyIntegrator)

    proj = net.connect(inp, pop, 'exc')
    proj.one_to_one(1.0)

    net.compile()

    net.simulate(10.)
    ```

    This creates a population of 10 neurons whose activity will change during the first 10*dt milliseconds of the simulation. After that delay, the last input will be kept (i.e. 1 for the last neuron).

    If you want the TimedArray to "loop" over the different input vectors, you can specify a period for the inputs:

    ```python
    inp = net.create(ann.TimedArray(rates=inputs, period=10.))
    ```

    If the period is smaller than the length of the rates, the last inputs will not be set.

    If you do not want the inputs to be set at every step, but every 10 ms for example, you can use the ``schedule`` argument:

    ```python
    inp = net.create(ann.TimedArray(rates=inputs, schedule=10.))
    ```

    The input [1, 0, 0,...] will stay for 10 ms, then [0, 1, 0, ...] for the next 10 ms, etc...

    If you need a less regular schedule, you can specify it as a list of times:

    ```python
    inp = ann.TimedArray(rates=inputs, schedule=[10., 20., 50., 60., 100., 110.])
    ```

    The first input is set at t = 10 ms (r = 0.0 in the first 10 ms), the second at t = 20 ms, the third at t = 50 ms, etc.

    If you specify less times than in the array of rates, the last ones will be ignored.

    Scheduling can be combined with periodic cycling. Note that you can use the ``reset()`` method to manually reinitialize the TimedArray, times becoming relative to that call:

    ```python
    ann.simulate(100.) # ten inputs are shown with a schedule of 10 ms
    inp.reset()
    ann.simulate(100.) # the same ten inputs are presented again.
    ```

    :param rates: array of firing rates. The first axis corresponds to time, the others to the desired dimensions of the population.
    :param geometry: desired dimensions of the population. This argument will be considered if *rates* is None.
    :param schedule: either a single value or a list of time points where inputs should be set. Default: every timestep.
    :param period: time when the timed array will be reset and start again, allowing cycling over the inputs. Default: no cycling (-1.).

    """
    def __init__(self, 
                 rates:np.ndarray=None, 
                 geometry:int|tuple=None, 
                 schedule:float=0., 
                 period:float=-1., 
                 name:str=None, 
                 copied:bool=False,
                 net_id:int=0,
                 ):

        # Sanity check
        if rates is None and geometry is None:
            Messages._error("TimedArray: either *rates* or *geometry* argument must be set.")

        # Geometry of the population
        if rates is not None:
            if geometry is None:
                geometry = rates.shape[1:]
            else:
                if geometry != rates.shape[1:]:
                    Messages._warning("TimedArray: mismatch between *rates* and *geometry* dimensions detected.")

        # Create input neuron type
        neuron = Neuron(
            parameters="",
            equations="r = 0.0",
            name="Timed Array",
            description="Timed array sets inputs (shape = {}) sequentially with schedule = {} and period = {}.".format(geometry, schedule, period)
        )

        SpecificPopulation.__init__(self, geometry=geometry, neuron=neuron, name=name, copied=copied, net_id=net_id)

        self.init['schedule'] = schedule
        self.init['rates'] = rates
        self.init['period'] = period

        if rates is not None:
            self.update(rates=rates, period=period, schedule=schedule)

    @property
    def r(self):
        if self.initialized:
            return self._get_cython_attribute("r")
        else:
            Messages._error("Read-out of 'r' is only possible after compile.")

    @r.setter
    def r(self, new_r):
        Messages._error("The value of r is defined through the '*'rates' argument.")

    def update(self, rates:np.ndarray, schedule:float=0., period:float=-1) -> None:
        """
        Set a new array of inputs. 
        
        The first axis corresponds to time, the others to the desired dimensions of the population. Note, the
        geometry is set during construction phase of the object.

        :param rates: array of firing rates. The first axis corresponds to time, the others to the desired dimensions of the population.
        :param schedule: either a single value or a list of time points where inputs should be set. Default: every timestep.
        :param period: time when the timed array will be reset and start again, allowing cycling over the inputs. Default: no cycling (-1.).
        """

        # Check the schedule
        if isinstance(schedule, (int, float)):
            if float(schedule) <= 0.0:
                schedule = ConfigManager().get('dt', self.net_id)

            self.schedule = [ float(schedule*i) for i in range(rates.shape[0])]
        else:
            self.schedule = schedule

        if len(self.schedule) > rates.shape[0]:
            Messages._error('TimedArray: the length of the schedule parameter cannot exceed the first dimension of the rates parameter.')

        if len(self.schedule) < rates.shape[0]:
            Messages._warning('TimedArray: the length of the schedule parameter is smaller than the first dimension of the rates parameter (more data than time points). Make sure it is what you expect.')

        self.rates = np.array(rates)
        self.period = period

    def _copy(self, net_id=None):
        "Returns a copy of the population when creating networks."
        return TimedArray(
            rates=self.rates, 
            geometry=self.geometry, 
            schedule=self.schedule, 
            period=self.period, 
            name=self.name, 
            copied=True, 
            net_id=self.net_id if net_id is None else net_id)

    def _generate_st(self):
        """
        adjust code templates for the specific population for single thread and openMP.
        """
        self._specific_template['declare_additional'] = """
    // Custom local parameters of a TimedArray
    std::vector< int > _schedule; // List of times where new inputs should be set
    std::vector< std::vector< %(float_prec)s > > _buffer; // buffer holding the data
    int _period; // Period of cycling
    long int _t; // Internal time
    int _block; // Internal block when inputs are set not at each step
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}
        
        self._specific_template['access_additional'] = """
    // Custom local parameters of a TimedArray
    void set_schedule(std::vector<int> schedule) { _schedule = schedule; }
    std::vector<int> get_schedule() { return _schedule; }
    void set_buffer(std::vector< std::vector< %(float_prec)s > > buffer) { _buffer = buffer; r = _buffer[_block]; }
    std::vector< std::vector< %(float_prec)s > > get_buffer() { return _buffer; }
    void set_period(int period) { _period = period; }
    int get_period() { return _period; }
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}
        
        self._specific_template['init_additional'] = """
        // Initialize counters
        _t = 0;
        _block = 0;
        _period = -1;
"""

        self._specific_template['reset_additional'] ="""
        _t = 0;
        _block = 0;

        r.clear();
        r = std::vector<%(float_prec)s>(size, 0.0);
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}


        self._specific_template['update_variables'] = """
        if(_active) {
        #ifdef _DEBUG
            std::cout << "TimedArray::update() - " << _t << " " << _block<< " " << _schedule[_block] << std::endl;
        #endif

            // Check if it is time to set the input
            if (_t == _schedule[_block]) {
                // sanity check
                if (_buffer.empty()) {
                    std::cerr << "TimedArray: no data being set ..." << std::endl;
                    r = std::vector<%(float_prec)s>(size, 0.0);
                    return;
                }

                // sanity check
                if (_buffer.size() <= _block) {
                    std::cerr << "TimedArray: not enough data being set ..." << std::endl;
                    r = std::vector<%(float_prec)s>(size, 0.0);
                    return;
                }

                // Set the data
                r = _buffer[_block];
                // Move to the next block
                _block++;
                // If was the last block, go back to the first block
                if (_block == _schedule.size()){
                    _block = 0;
                }
            }

            // If the timedarray is periodic, check if we arrive at that point
            if(_period > -1 && (_t == _period-1)){
                // Reset the counters
                _block=0;
                _t = -1;
            }

            // Always increment the internal time
            _t++;

        #ifdef _DEBUG
            std::cout << "TimedArray::update(t="<< t <<") - current buffer (min/max) = [" << *std::min_element(r.begin(), r.end()) << "," << *std::max_element(r.begin(), r.end()) <<  "]" << std::endl;
        #endif
        }
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}

        self._specific_template['size_in_bytes'] = """
        // schedule
        size_in_bytes += _schedule.capacity() * sizeof(int);

        // buffer
        size_in_bytes += _buffer.capacity() * sizeof(std::vector<%(float_prec)s>);
        for( auto it = _buffer.begin(); it != _buffer.end(); it++ )
            size_in_bytes += it->capacity() * sizeof(%(float_prec)s);
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}
        
        # Nanobind
        self._specific_template['wrapper'] = f"""
    // TimedArray
    nanobind::class_<PopStruct{self.id}>(m, "pop{self.id}_wrapper")
        // Constructor
        .def(nanobind::init<int, int>())

        // Common attributes
        .def_rw("size", &PopStruct{self.id}::size)
        .def_rw("max_delay", &PopStruct{self.id}::max_delay)

        // Attributes
		.def_rw("r", &PopStruct{self.id}::r)

        // Access methods
        .def("set_schedule", &PopStruct{self.id}::set_schedule)
        .def("get_schedule", &PopStruct{self.id}::get_schedule)

        .def("set_rates", &PopStruct{self.id}::set_buffer)
        .def("get_rates", &PopStruct{self.id}::get_buffer)

        .def("set_period", &PopStruct{self.id}::set_period)
        .def("get_period", &PopStruct{self.id}::get_period)

        // Other methods

        .def("activate", &PopStruct{self.id}::set_active)
        .def("reset", &PopStruct{self.id}::reset)
        .def("clear", &PopStruct{self.id}::clear);

"""

    def _generate_omp(self):
        """
        adjust code templates for the specific population for single thread and openMP.
        """
        self._specific_template['declare_additional'] = """
    // Custom local parameters of a TimedArray
    std::vector< int > _schedule; // List of times where new inputs should be set
    std::vector< std::vector< %(float_prec)s > > _buffer; // buffer holding the data
    int _period; // Period of cycling
    long int _t; // Internal time
    int _block; // Internal block when inputs are set not at each step
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}
        
        self._specific_template['access_additional'] = """
    // Custom local parameters of a TimedArray
    void set_schedule(std::vector<int> schedule) { _schedule = schedule; }
    std::vector<int> get_schedule() { return _schedule; }
    void set_buffer(std::vector< std::vector< %(float_prec)s > > buffer) { _buffer = buffer; r = _buffer[_block]; }
    std::vector< std::vector< %(float_prec)s > > get_buffer() { return _buffer; }
    void set_period(int period) { _period = period; }
    int get_period() { return _period; }
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}
        
        self._specific_template['init_additional'] = """
        // Initialize counters
        _t = 0;
        _block = 0;
        _period = -1;
"""
        
        self._specific_template['reset_additional'] ="""
        _t = 0;
        _block = 0;

        r.clear();
        r = std::vector<%(float_prec)s>(size, 0.0);
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}


        # HD (28th Jun. 24): contrary to the single-thread codes, where we use 'return' to escape the function
        #                    execution, OpenMP does not allow 'return'/'continue' in the execution block. Therefore,
        #                    we need to use the if-else tree.
        self._specific_template['update_variables'] = """
        if(_active){
            #pragma omp single
            {
            #ifdef _DEBUG
                std::cout << "TimedArray::update() - " << _t << " " << _block<< " " << _schedule[_block] << std::endl;
            #endif

                // Check if it is time to set the input
                if (_t == _schedule[_block]) {
                    // sanity check
                    if (_buffer.empty()) {
                        std::cerr << "TimedArray: no data being set ..." << std::endl;
                        r = std::vector<%(float_prec)s>(size, 0.0);
                    }

                    // sanity check
                    else if (_buffer.size() <= _block) {
                        std::cerr << "TimedArray: not enough data being set ..." << std::endl;
                        r = std::vector<%(float_prec)s>(size, 0.0);
                    }

                    // everything appears right, proceed
                    else {
                        // Set the data
                        r = _buffer[_block];
                        // Move to the next block
                        _block++;
                        // If was the last block, go back to the first block
                        if (_block == _schedule.size()){
                            _block = 0;
                        }
                    }
                }

                // If the timedarray is periodic, check if we arrive at that point
                if(_period > -1 && (_t == _period-1)){
                    // Reset the counters
                    _block=0;
                    _t = -1;
                }

                // Always increment the internal time
                _t++;
            }
        }
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}

        self._specific_template['size_in_bytes'] = """
        // schedule
        size_in_bytes += _schedule.capacity() * sizeof(int);

        // buffer
        size_in_bytes += _buffer.capacity() * sizeof(std::vector<%(float_prec)s>);
        for( auto it = _buffer.begin(); it != _buffer.end(); it++ )
            size_in_bytes += it->capacity() * sizeof(%(float_prec)s);
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}
        

        # Nanobind
        self._specific_template['wrapper'] = f"""
    // TimedArray
    nanobind::class_<PopStruct{self.id}>(m, "pop{self.id}_wrapper")
        // Constructor
        .def(nanobind::init<int, int>())

        // Common attributes
        .def_rw("size", &PopStruct{self.id}::size)
        .def_rw("max_delay", &PopStruct{self.id}::max_delay)

        // Attributes
		.def_rw("r", &PopStruct{self.id}::r)

        // Access methods
        .def("set_schedule", &PopStruct{self.id}::set_schedule)
        .def("get_schedule", &PopStruct{self.id}::get_schedule)

        .def("set_rates", &PopStruct{self.id}::set_buffer)
        .def("get_rates", &PopStruct{self.id}::get_buffer)

        .def("set_period", &PopStruct{self.id}::set_period)
        .def("get_period", &PopStruct{self.id}::get_period)

        // Other methods

        .def("activate", &PopStruct{self.id}::set_active)
        .def("reset", &PopStruct{self.id}::reset)
        .def("clear", &PopStruct{self.id}::clear);

"""

    def _generate_cuda(self):
        """
        adjust code templates for the specific population for single thread and CUDA.
        """
        # HD (18. Nov 2016)
        # I suppress the code generation for allocating the variable r on gpu, as
        # well as memory transfer codes. This is only possible as no other variables
        # allowed in TimedArray.
        self._specific_template['init_parameters_variables'] = ""
        self._specific_template['host_device_transfer'] = ""
        self._specific_template['device_host_transfer'] = ""

        #
        # Code for handling the buffer and schedule parameters
        self._specific_template['declare_additional'] = """
    // Custom local parameter timed array
    std::vector< int > _schedule;
    std::vector< %(float_prec)s* > _gpu_buffer;
    int _period; // Period of cycling
    long int _t; // Internal time
    int _block; // Internal block when inputs are set not at each step
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}
        
        self._specific_template['access_additional'] = """
    // Custom local parameter timed array
    void set_schedule(std::vector<int> schedule) { _schedule = schedule; }
    std::vector<int> get_schedule() { return _schedule; }
    void set_buffer(std::vector< std::vector< %(float_prec)s > > buffer) {
    #ifdef _DEBUG
        std::cout << "PopStruct%(id)s::set_buffer()" << std::endl;
    #endif
        // clear a previous allocated container.
        if ( !_gpu_buffer.empty() ) {
            for (auto it = _gpu_buffer.begin(); it != _gpu_buffer.begin(); it++) {
                cudaFree(*it);
            }
            _gpu_buffer.clear();
            _gpu_buffer.shrink_to_fit();
        }
        // abort updating the container if no data is provided.
        if (buffer.empty()) {
            std::cerr << "The buffer provided to TimedArray should not be empty!" << std::endl;
            gpu_r = nullptr;
            _gpu_buffer = _gpu_buffer = std::vector< %(float_prec)s* >();
            return;
        }

        // host holds a set of pointers
        _gpu_buffer = std::vector< %(float_prec)s* >(buffer.size(), nullptr);

        // allocate gpu arrays
        for(int i = 0; i < buffer.size(); i++) {
            cudaMalloc((void**)&_gpu_buffer[i], buffer[i].size()*sizeof(%(float_prec)s));
        }

        auto host_it = buffer.begin();
        auto dev_it = _gpu_buffer.begin();
        for (; host_it != buffer.end(); host_it++, dev_it++) {
            cudaMemcpy( *dev_it, host_it->data(), host_it->size()*sizeof(%(float_prec)s), cudaMemcpyHostToDevice);
        }

        gpu_r = _gpu_buffer[_block];
    }
    std::vector< std::vector< %(float_prec)s > > get_buffer() {
        std::vector< std::vector< %(float_prec)s > > buffer = std::vector< std::vector< %(float_prec)s > >( _gpu_buffer.size(), std::vector<%(float_prec)s>(size,0.0) );

        auto host_it = buffer.begin();
        auto dev_it = _gpu_buffer.begin();
        for (; host_it < buffer.end(); host_it++, dev_it++) {
            cudaMemcpy( host_it->data(), *dev_it, size*sizeof(%(float_prec)s), cudaMemcpyDeviceToHost );
        }

        return buffer;
    }
    void set_period(int period) { _period = period; }
    int get_period() { return _period; }
""" % {'id': self.id, 'float_prec': ConfigManager().get('precision', self.net_id)}
        
        self._specific_template['init_additional'] = """
        // counters
        _t = 0;
        _block = 0;
        _period = -1;
"""

        self._specific_template['reset_additional'] = """
        // counters
        _t = 0;
        _block = 0;
        if (!_gpu_buffer.empty())
            gpu_r = _gpu_buffer[0];
"""

        # there is no GPU-side computation
        self._specific_template['update_variable_body'] = ""
        self._specific_template['update_variable_invoke'] = ""
        self._specific_template['update_variable_header'] = ""

        # we switch the GPU buffer which is read out in each time step
        self._specific_template['update_variables'] = """
        if(_active) {
        #ifdef _DEBUG
            std::cout << "TimedArray::update() - " << _t << " " << _block<< " " << _schedule[_block] << std::endl;
        #endif
            // Check if it is time to set the input
            if (_t == _schedule[_block]) {
                // sanity check
                if (_gpu_buffer.empty()) {
                    std::cerr << "TimedArray: no data being set ..." << std::endl;
                    gpu_r = _gpu_buffer[0];
                    return;
                }

                // sanity check
                if (_gpu_buffer.size() <= _block) {
                    std::cerr << "TimedArray: not enough data being set ..." << std::endl;
                    gpu_r = _gpu_buffer[0];
                    return;
                }

                // Set the data
                gpu_r = _gpu_buffer[_block];
                // Move to the next block
                _block++;
                // If was the last block, go back to the first block
                if ( _block == _schedule.size() ) {
                    _block = 0;
                }
            }

            // If the timedarray is periodic, check if we arrive at that point
            if( (_period > -1) && (_t == _period-1) ) {
                // Reset the counters
                _block=0;
                _t = -1;
            }

            // Always increment the internal time
            _t++;
        }
"""
        # call the switch of CPU-buffers (host-side)
        self._specific_template['update_variable_call'] = """
    // host side update of neurons
    pop%(id)s->update();
""" % {'id': self.id}

        self._specific_template['size_in_bytes'] = """
        // r
        size_in_bytes += sizeof(std::vector<%(float_prec)s>);
        size_in_bytes += r.capacity() * sizeof(%(float_prec)s);

        // schedule
        size_in_bytes += sizeof(std::vector<int>);
        size_in_bytes += _schedule.capacity() * sizeof(int);

        // gpu_buffer
        size_in_bytes += sizeof(std::vector<%(float_prec)s*>);
        size_in_bytes += _gpu_buffer.capacity() * sizeof(%(float_prec)s*);
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}
        

        # Nanobind
        self._specific_template['wrapper'] = f"""
    // TimedArray
    nanobind::class_<PopStruct{self.id}>(m, "pop{self.id}_wrapper")
        // Constructor
        .def(nanobind::init<int, int>())

        // Common attributes
        .def_rw("size", &PopStruct{self.id}::size)
        .def_rw("max_delay", &PopStruct{self.id}::max_delay)

        // Attributes
		.def_rw("r", &PopStruct{self.id}::r)
        .def_rw("r_host_to_device", &PopStruct{self.id}::r_host_to_device)

        // Access methods
        .def("set_schedule", &PopStruct{self.id}::set_schedule)
        .def("get_schedule", &PopStruct{self.id}::get_schedule)

        .def("set_rates", &PopStruct{self.id}::set_buffer)
        .def("get_rates", &PopStruct{self.id}::get_buffer)

        .def("set_period", &PopStruct{self.id}::set_period)
        .def("get_period", &PopStruct{self.id}::get_period)

        // Other methods

        .def("activate", &PopStruct{self.id}::set_active)
        .def("reset", &PopStruct{self.id}::reset)
        .def("clear", &PopStruct{self.id}::clear);

"""

    def _instantiate(self, module):
        # Create the Cython instance
        self.cyInstance = getattr(module, self.class_name+'_wrapper')(self.size, self.max_delay)

    def __setattr__(self, name, value):
        if name == 'schedule':
            if self.initialized:
                val_int = np.array((np.atleast_1d(value) / ConfigManager().get('dt', self.net_id)), dtype=np.int32)
                self.cyInstance.set_schedule( val_int )
            else:
                self.init['schedule'] = value
        elif name == 'rates':
            if self.initialized:
                if value is None:
                    return # nothing to do

                if len(value.shape) > 2:
                    if value.shape[1:] != self.geometry:
                        Messages._error("TimedArray: mismatch between *rates* argument (", value.shape[1:], ") and stored geometry (", self.geometry, ").")

                    # we need to flatten the provided data
                    flat_values = value.reshape( (value.shape[0], self.size) )
                    self.cyInstance.set_rates( flat_values )
                else:
                    self.cyInstance.set_rates( value )
            else:
                self.init['rates'] = value
        elif name == "period":
            if self.initialized:
                self.cyInstance.set_period(int(value /ConfigManager().get('dt', self.net_id)))
            else:
                self.init['period'] = value
        else:
            Population.__setattr__(self, name, value)

    def __getattr__(self, name):
        if name == 'schedule':
            if self.initialized:
                return [ConfigManager().get('dt', self.net_id) * val for val in self.cyInstance.get_schedule()]
            else:
                return self.init['schedule']
        elif name == 'rates':
            if self.initialized:
                if len(self.geometry) > 1:
                    # unflatten the data
                    flat_values = self.cyInstance.get_rates()
                    values = np.zeros( tuple( [len(self.schedule)] + list(self.geometry) ) )
                    for x in range(len(self.schedule)):
                        values[x] = np.reshape( flat_values[x], self.geometry)
                    return values
                else:
                    return self.cyInstance.get_rates()
            else:
                return self.init['rates']
        elif name == 'period':
            if self.initialized:
                return self.cyInstance.get_period() * ConfigManager().get('dt', self.net_id)
            else:
                return self.init['period']
        else:
            return Population.__getattribute__(self, name)
        




"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from sympy import Symbol, Function
from ANNarchy.intern.GlobalObjects import GlobalObjectManager
from ANNarchy.intern import Messages

# Dictionary of default elements for the C++ generation
parser_dict = {
    'dt' : Symbol('dt'),
    't' : Symbol('(double(t)*dt)'),
    'w' : Symbol('w%(local_index)s'),
    'g_target': Symbol('sum'), # TODO: still useful?
    't_last': Symbol('((double)(last_spike%(local_index)s)*dt)'),
    't_pre': Symbol('((double)(%(pre_prefix)slast_spike%(pre_index)s)*dt)'),
    't_post': Symbol('((double)(%(post_prefix)slast_spike%(post_index)s)*dt)')
}

# Dictionary of built-in functions for the C++ generation
functions_dict = {
    'pos': Function('positive'),
    'positive': Function('positive'),
    'neg': Function('negative'),
    'negative': Function('negative'),
    'modulo': Function('modulo'),
    'fabs': Function('fabs'),
    'ite': Function('ite', nargs=3),
    'clip': Function('clip'),
    'True': Symbol('true'),
    'False': Symbol('false'),
    'power': Function('power', nargs=2),
    'round': Function('round')
}

# Built-in functions with their correct name
user_functions = {
    'pos': 'positive',
    'positive': 'positive',
    'neg': 'negative',
    'negative': 'negative',
    'modulo': 'modulo',
    'fabs': 'fabs',
    'clip': 'clip',
    'ite': 'ite',
    'power': 'power',
    'isfinite': 'std::isfinite',
    'isnan': 'std::isnan'
}

def create_local_dict(local_attributes, semiglobal_attributes, global_attributes, untouched, local_functions=[]):
    """
    Creates a dictionary of Sympy symbols based on the attributes of the neuron/synapse.

    *Parameters:*

    * `local_attributes`: local attributes.
    * `semiglobal_attributes`: semiglobal attributes (postsynaptic).
    * `global_attributes`: global attributes.
    * `untouched`: list of variable names which were replaced (e.g. random distributions, targets)

    """

    # Copy the default dictionary of built-in symbols or functions
    local_dict = parser_dict.copy()
    local_dict.update(functions_dict)

    # Add each variable of the neuron depending on its locality
    for var in local_attributes:
        if var in functions_dict.keys():
            Messages._error(var, "is a reserved keyword in ANNarchy.")
        local_dict[var] = Symbol(var + '%(local_index)s')
    for var in semiglobal_attributes:
        if var in functions_dict.keys():
            Messages._error(var, "is a reserved keyword in ANNarchy.")
        local_dict[var] = Symbol(var + '%(semiglobal_index)s')
    for var in global_attributes:
        if var in functions_dict.keys():
            Messages._error(var, "is a reserved keyword in ANNarchy.")
        local_dict[var] = Symbol(var + '%(global_index)s')

    # Add custom constants
    for name in GlobalObjectManager().list_constants():
        if name in local_dict.keys():
            continue
        local_dict[name] = Symbol(name)

    # Add each untouched variable
    for var in untouched:
        local_dict[var] = Symbol(var)

    # Add local functions
    for var in local_functions:
        local_dict[var] = Function(var)

    return local_dict



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

pop_struct_wrapper = """
    // PopStruct%(id)s
    nanobind::class_<PopStruct%(id)s>(m, "pop%(id)s_wrapper")
        // Constructor
        .def(nanobind::init<int, int>())

        // Common attributes
        .def_rw("size", &PopStruct%(id)s::size)
        .def_rw("max_delay", &PopStruct%(id)s::max_delay)

        // Attributes
%(attributes)s

        // Other methods
%(additional)s
        .def("activate", &PopStruct%(id)s::set_active)
        .def("reset", &PopStruct%(id)s::reset)
        .def("clear", &PopStruct%(id)s::clear);
"""

pop_mon_wrapper = """
    // Monitor for Population %(id)s
    nanobind::class_<PopRecorder%(id)s>(m, "PopRecorder%(id)s_wrapper")
        // Record flag
%(record_flag)s

        // Target container
%(record_container)s

        // Clear container
%(clear_container)s

        // Functions
        .def(nanobind::init<std::vector<int>, int, int, long>())
        .def("clear", &PopRecorder%(id)s::clear)
        .def("size_in_bytes", &PopRecorder%(id)s::size_in_bytes);
"""



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

proj_struct_wrapper = """
    // ProjStruct%(id)s
    nanobind::class_<ProjStruct%(id)s>(m, "proj%(id)s_wrapper")
        // Constructor
        .def(nanobind::init<>())

        // Flags
        .def_rw("_transmission", &ProjStruct%(id)s::_transmission)
        .def_rw("_axon_transmission", &ProjStruct%(id)s::_axon_transmission)
        .def_rw("_update", &ProjStruct%(id)s::_update)
        .def_rw("_update_period", &ProjStruct%(id)s::_update_period)
        .def_rw("_update_offset", &ProjStruct%(id)s::_update_offset)
        .def_rw("_plasticity", &ProjStruct%(id)s::_plasticity)

        // Connectivity
%(connectivity)s

        // Methods
%(methods)s

        // Attributes
%(attributes)s

        // Other methods
%(additional)s
        .def("size_in_bytes", &ProjStruct%(id)s::size_in_bytes)
        .def("clear", &ProjStruct%(id)s::clear);
"""

proj_mon_wrapper = """
    // Monitor for Projection %(id)s
    nanobind::class_<ProjRecorder%(id)s>(m, "ProjRecorder%(id)s_wrapper")
        // Record flag
%(record_flag)s

        // Target container
%(record_container)s

        // Clear container
%(clear_container)s

        // Functions
        .def(nanobind::init<std::vector<int>, int, int, long>())
        .def("clear", &ProjRecorder%(id)s::clear)
        .def("size_in_bytes", &ProjRecorder%(id)s::size_in_bytes);
"""

proj_local_attr = """
        // local attributes
        .def("get_local_attribute_all_%(ctype)s", &ProjStruct%(id)s::get_local_attribute_all_%(ctype)s)
        .def("get_local_attribute_row_%(ctype)s", &ProjStruct%(id)s::get_local_attribute_row_%(ctype)s)
        .def("get_local_attribute_%(ctype)s", &ProjStruct%(id)s::get_local_attribute_%(ctype)s)

        .def("set_local_attribute_all_%(ctype)s", &ProjStruct%(id)s::set_local_attribute_all_%(ctype)s)
        .def("set_local_attribute_row_%(ctype)s", &ProjStruct%(id)s::set_local_attribute_row_%(ctype)s)
        .def("set_local_attribute_%(ctype)s", &ProjStruct%(id)s::set_local_attribute_%(ctype)s)
"""

proj_semiglobal_attr = """
        // semiglobal attributes
        .def("get_semiglobal_attribute_all_%(ctype)s", &ProjStruct%(id)s::get_semiglobal_attribute_all_%(ctype)s)
        .def("get_semiglobal_attribute_%(ctype)s", &ProjStruct%(id)s::get_semiglobal_attribute_%(ctype)s)

        .def("set_semiglobal_attribute_all_%(ctype)s", &ProjStruct%(id)s::set_semiglobal_attribute_all_%(ctype)s)
        .def("set_semiglobal_attribute_%(ctype)s", &ProjStruct%(id)s::set_semiglobal_attribute_%(ctype)s)
"""

proj_global_attr = """
        // global attributes
        .def("get_global_attribute_%(ctype)s", &ProjStruct%(id)s::get_global_attribute_%(ctype)s)
        .def("set_global_attribute_%(ctype)s", &ProjStruct%(id)s::set_global_attribute_%(ctype)s)
"""

proj_delays = {
    'uniform': """
        // Synaptic delays
        .def("get_delay", &ProjStruct%(id)s::get_delay)
        .def("get_dendrite_delay", &ProjStruct%(id)s::get_dendrite_delay)
        .def("set_delay", &ProjStruct%(id)s::set_delay)
""",
    'nonuniform_rate_coded': """
        // Synaptic delays
        .def("get_delay", &ProjStruct%(id)s::get_delay)
        .def("get_dendrite_delay", &ProjStruct%(id)s::get_dendrite_delay)
        .def("set_delay", &ProjStruct%(id)s::set_delay)

        .def("get_max_delay", &ProjStruct%(id)s::get_max_delay)
        .def("set_max_delay", &ProjStruct%(id)s::set_max_delay)
""",
    'nonuniform_spiking': """
        // Synaptic delays
        .def("get_delay", &ProjStruct%(id)s::get_delay)
        .def("get_dendrite_delay", &ProjStruct%(id)s::get_dendrite_delay)
        .def("set_delay", &ProjStruct%(id)s::set_delay)

        .def("get_max_delay", &ProjStruct%(id)s::get_max_delay)
        .def("set_max_delay", &ProjStruct%(id)s::set_max_delay)

        .def("update_max_delay", &ProjStruct%(id)s::update_max_delay)
        .def("reset_ring_buffer", &ProjStruct%(id)s::reset_ring_buffer)
"""
}